import importlib
import socket
import time
import re
from threading import Lock

import numpy as np

from empyric.tools import read_from_socket, write_to_socket


def chaperone(method):
    """
    Wraps all write, read and query methods of the adapters; monitors and
    handles communication issues.

    :param method: (callable) method to be wrapped
    :return: (callable) wrapped method
    """

    def wrapped_method(self, *args, validator=None, **kwargs):
        if not self.connected:
            raise AdapterError(
                "Adapter is not connected for instrument "
                f"at address {self.instrument.address}"
            )

        # Catch communication errors and either try to repeat communication
        # or reset the connection
        attempts = 0
        reconnects = 0

        self.lock.acquire()

        while reconnects <= self.max_reconnects:
            while attempts < self.max_attempts:
                try:
                    response = method(self, *args, **kwargs)

                    if validator and not validator(response):
                        if hasattr(response, "__len__") and len(response) > 100:
                            response = str(response[:50]) + "..." + str(response[-50:])

                        raise ValueError(
                            f"invalid response, {response}, "
                            f"from {method.__name__} method"
                        )

                    elif attempts > 0 or reconnects > 0:
                        print("Resolved")

                    self.lock.release()
                    return response

                except BaseException as err:
                    print(
                        f"Encountered {err} while trying "
                        f"to talk to {self.instrument.name}"
                        "\nRetrying..."
                    )
                    attempts += 1

            # repeats have maxed out, so try reconnecting with the instrument
            print("Reconnecting...")
            self.disconnect()
            time.sleep(self.delay)
            self.connect()

            attempts = 0
            reconnects += 1

        # Getting here means that both repeats
        # and reconnects have been maxed out
        self.lock.release()
        raise AdapterError(f"Unable to communicate with {self.instrument.name}!")

    wrapped_method.__doc__ = method.__doc__  # keep method doc string

    return wrapped_method


class AdapterError(BaseException):
    pass


class Adapter:
    """
    Base class for all adapters
    """

    #: Maximum number of attempts to read from a port/channel,
    # in the event of a communication error
    max_attempts = 3

    #: Maximum number of times to try to reset communications,
    # in the event of a communication error
    max_reconnects = 1

    kwargs = []

    # Library used by adapter; overwritten in children classes.
    lib = "python"

    # If upon instantiation no valid library is found for adapter, raise
    # AdapterError with the following message; overwritten in children classes.
    no_lib_msg = "no valid library found for adapter; " "check library installation"

    delay = 0.1  # delay between successive communication attempts

    def __init__(self, instrument, **kwargs):
        if self.lib is None:
            # determined by class attribute `lib`
            raise AdapterError(self.no_lib_msg)

        # general parameters
        self.instrument = instrument
        self.backend = None
        self.connected = False
        self.repeats = 0
        self.reconnects = 0

        for key, value in kwargs.items():
            self.__setattr__(key, value)

        self.connect()

        self.instrument.adapter = self

        # This lock is used by the chaperone wrapper function to prevent
        # cross-talk from different threads. Each time an adapter's write, read
        # or query methods is called, the lock is acquired and then released
        # when the transaction is complete.
        self.lock = Lock()

    def __del__(self):
        # Try to cleanly close communications when adapters are deleted
        if self.connected:
            try:
                self.disconnect()
            except BaseException as err:
                print(f"Error while disconnecting {self.instrument.name}:", err)
                pass

    # All methods below should be overwritten in child class definitions

    def __repr__(self):
        return "Adapter"

    def connect(self):
        """
        Establishes communications with the instrument through the appropriate
        backend.
        """
        self.connected = True

    @chaperone
    def write(self, *args, validator=None, **kwargs):
        """
        Write a command.

        :param args: any arguments for the write method
        :param validator: (callable) function that returns True if its input
                          looks right or False if it does not
        :param kwargs: any keyword arguments for the write method

        :return: (str) literal 'Success' if write operation is successful
        """

        if hasattr(self, "_write"):
            return self._write(*args, **kwargs)
        else:
            raise AttributeError(self.__name__ + " adapter has no _write method")

    @chaperone
    def read(self, *args, validator=None, **kwargs):
        """
        Read an awaiting message.

        :param args: any arguments for the read method
        :param validator: (callable) function that returns True if its input
                          looks right or False if it does not
        :param kwargs: any keyword arguments for the read method

        :return: instrument response
        """

        if hasattr(self, "_read"):
            return self._read(*args, **kwargs)
        else:
            raise AttributeError(self.__name__ + " adapter has no _read method")

    @chaperone
    def query(self, *args, validator=None, **kwargs):
        """
        Submit a query.

        :param args: any arguments for the query method
        :param validator: (callable) function that returns True if its input
                          looks right or False if it does not
        :param kwargs: any keyword arguments for the query method

        :return: instrument response
        """

        if hasattr(self, "_query"):
            return self._query(*args, **kwargs)
        else:
            raise AttributeError(self.__name__ + " adapter has no _query method")

    def disconnect(self):
        """
        Close communication port/channel
        """
        self.connected = False


class Serial(Adapter):
    """
    Handles communications with serial instruments through either PyVISA or
    PySerial. If both are installed, it defaults to PyVISA.
    """

    baud_rate = 9600
    timeout = 0.1
    delay = 0.1
    parity = "N"
    stop_bits = 1
    read_termination = "\n"
    write_termination = "\r"

    kwargs = [
        "baud_rate",
        "timeout",
        "delay",
        "byte_size",
        "parity",
        "stop_bits",
    ]

    # Get serial library
    if importlib.util.find_spec("pyvisa"):
        lib = "pyvisa"
    elif importlib.util.find_spec("serial"):
        lib = "pyserial"
    else:
        lib = None

    no_lib_msg = (
        "No serial library was found! " "Please install either PySerial or PyVISA."
    )

    def connect(self):
        # First try connecting with PyVISA
        if self.lib == "pyvisa":
            pyvisa = importlib.import_module("pyvisa")

            self.backend = pyvisa.open_resource(
                self.instrument.address,
                baud_rate=self.baud_rate,
                stop_bits=self.stop_bits,
                parity=self.parity,
                timeout=self.timeout,
                write_termination=self.write_termination,
                read_terimation=self.read_termination,
            )

        # Then try connecting with PySerial
        elif self.lib == "pyserial":
            serial = importlib.import_module("serial")

            self.backend = serial.Serial(
                port=self.instrument.address,
                baudrate=self.baud_rate,
                stopbits=self.stop_bits,
                parity=self.parity,
                timeout=self.timeout,
            )

        else:
            raise AdapterError(f"invalid library specification, {self.lib}")

        self.connected = True

    def _write(self, message):
        if self.lib == "pyvisa":
            self.backend.write(message)
        elif self.lib == "pyserial":
            self.backend.write((message + self.write_termination).encode())

        return "Success"

    def _read(self, bytes=None, until=None, decode=True):
        if self.lib == "pyvisa":
            if bytes:
                response = self.backend.read_bytes(bytes)
            elif until:
                response = b""
                while until.encode() not in response:
                    response = response + self.backend.read_raw(1)
            else:
                return self.backend.read(decode=False)  # decoded below

        elif self.lib == "pyserial":
            if bytes:
                response = self.backend.read(bytes)
            elif until:
                response = self.backend.read_until(until)
            else:
                response = self.backend.read_until(self.read_termination.encode())

        else:
            response = b""

        if decode:
            response = response.decode().strip()

        return response

    def _query(self, question, bytes=None, until=None, decode=True):
        self._write(question)
        time.sleep(self.delay)
        return self._read(bytes=bytes, until=until, decode=decode)

    def disconnect(self):
        if self.lib == "pyvisa":
            self.backend.clear()

        elif self.lib == "pyserial":
            self.backend.reset_input_buffer()
            self.backend.reset_output_buffer()

        self.backend.close()

        self.connected = False

    def __repr__(self):
        return "Serial"

    @classmethod
    def list(cls):
        """
        List all connected serial devices

        :param verbose: (bool) if True, list of devices will be printed
                        (defaults to True).

        :return: (list of str) List of connected serial devices
        """

        if cls.lib == "pyvisa":
            pyvisa = importlib.import_module("pyvisa")
            resource_manager = pyvisa.ResourceManager()

            devices = resource_manager.list_resources()

        elif cls.lib == "pyserial":
            list_ports = importlib.import_module("serial.tools.list_ports").comports

            devices = [port.device for port in list_ports()]

        else:
            raise AdapterError(cls.no_lib_msg)

        return devices

    @classmethod
    def locate(cls):
        """
        Determine the address of a serial instrument via the
        "unplug-it-then-plug-it-back-in" method.

        :return: None (address is printed to console)
        """

        input("Press enter when the instrument is disconnected")

        other_devices = Serial.list()

        input("Press enter when the instrument is connected")

        all_devices = Serial.list()

        try:
            instrument_address = [
                device for device in all_devices if device not in other_devices
            ][0]

            print(f"Address: {instrument_address}\n")

        except IndexError:
            print("Instrument not found!\n")

        again = "y" in input("Try again? [y/n]").lower()
        if again:
            Serial.locate()


class GPIB(Adapter):
    """
    Handles communications with GPIB instruments through either PyVISA,
    LinuxGPIB (if OS is Linux) for most GPIB-USB controllers (defaults to
    PyVISA, if installed). For Prologix GPIB-USB adapter units, this adapter
    uses PySerial to facilitate communications.
    """

    # Enumerated timeout values (in seconds) allowed by the Linux-GPIB backend
    linux_gpib_timeouts = {
        0: None,
        1: 10e-6,
        2: 30e-6,
        3: 100e-6,
        4: 300e-6,
        5: 1e-3,
        6: 3e-3,
        7: 10e-3,
        8: 30e-3,
        9: 100e-3,
        10: 300e-3,
        11: 1,
        12: 3,
        13: 10,
        14: 30,
        15: 100,
        16: 300,
        17: 1000,
    }

    kwargs = ["prologix_address"]

    prologix_address = None
    prologix_controllers = {}

    delay = 0.05
    _timeout = None

    # Get GPIB library
    if importlib.util.find_spec("pyvisa"):
        lib = "pyvisa"
    elif importlib.util.find_spec("gpib_ctypes"):
        lib = "linux-gpib"
    else:
        if importlib.util.find_spec("serial"):
            lib = "prologix-gpib"

    no_lib_msg = (
        "No valid library found for GPIB adapters!"
        "Please install PyVISA (with GPIB drivers), Linux-GPIB, or"
        "use a Prologix GPIB-USB or GPIB-ETHERNET adapter "
        "(requires PySerial)"
    )

    def __init__(self, instrument, **kwargs):
        super().__init__(instrument, **kwargs)
        self._descr = None

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        if self.connected:
            if self.lib == "pyvisa" or self.lib == "prologix-gpib":
                # pyvisa records timeouts in milliseconds
                if timeout is None:
                    self.backend.timeout = None
                else:
                    self.backend.timeout = timeout * 1000
                self._timeout = timeout
            elif self.lib == "linux-gpib":
                self._timeout = self._linux_gpib_set_timeout(timeout)
        else:
            self._timeout = None

    def connect(self):
        if self.prologix_address is not None:
            self.lib = "prologix-gpib"

        if self.lib == "pyvisa":
            visa = importlib.import_module("pyvisa")

            manager = visa.ResourceManager()

            full_address = None
            for address in manager.list_resources():
                instrument_address = str(self.instrument.address)

                address_format = "GPIB[0-9]::" + instrument_address + "::INSTR"

                address_match = re.match(address_format, address)

                if address_match:
                    full_address = address

            if full_address:
                self.backend = manager.open_resource(full_address)
            else:
                AdapterError(
                    "GPIB device at address " f"{self.instrument.address} not found!"
                )

        elif self.lib == "linux-gpib":
            self.backend = importlib.import_module("gpib")

            self._descr = self.backend.dev(0, self.instrument.address, 0, 9, 1, 0)

        elif self.lib == "prologix-gpib":
            if self.prologix_address is None:
                raise AdapterError(
                    "trying to connect to Prologix GPIB adapter but no address "
                    "was provided (prologix_address argument was not set); "
                    "must be either an IP address (for Prologix GPIB-LAN) or "
                    "serial port (for Prologix GPIB-USB)"
                )

            if self.prologix_address in GPIB.prologix_controllers:
                self.backend = GPIB.prologix_controllers[self.prologix_address]
            else:
                if re.match("\d+\.\d+\.\d+\.\d+", self.prologix_address):
                    self.backend = PrologixGPIBLAN(self.prologix_address)
                else:
                    self.backend = PrologixGPIBUSB(self.prologix_address)

                GPIB.prologix_controllers[self.prologix_address] = self.backend

            if self.instrument.address not in self.backend.devices:
                self.backend.devices.append(self.instrument.address)

        else:
            raise AdapterError(
                f"invalid library specification; options are 'pyvisa', "
                f"'linux-gpib' or 'prologix-gpib'. If using a Prologix GPIB "
                "adapter, its 'prologix address' argument must be specified"
            )

        self.connected = True

    def _write(self, message):
        if self.lib == "pyvisa":
            self.backend.write(message)
        elif self.lib == "linux-gpib":
            self.backend.write(self._descr, message)
        elif self.lib == "prologix-gpib":
            self.backend.write(message, address=self.instrument.address)

        return "Success"

    def _read(self, bytes=1024):
        if self.lib == "pyvisa":
            return self.backend.read()
        elif self.lib == "linux-gpib":
            return self.backend.read(self._descr, bytes).decode()
        elif self.lib == "prologix-gpib":
            return self.backend.read(address=self.instrument.address)

    def _query(self, question):
        self._write(question)
        time.sleep(self.delay)
        return self._read()

    def _linux_gpib_set_timeout(self, timeout):
        if timeout is None:
            self.backend.timeout(self.descr, 0)
            return None
        else:
            for index, allowed_timeout in self.linux_gpib_timeouts.items()[1:]:
                if allowed_timeout >= timeout:
                    self.backend.timeout(self.descr, index)
                    break

            return allowed_timeout

    def disconnect(self):
        if self.lib == "pyvisa":
            self.backend.clear()
            self.backend.close()
        elif self.lib == "linux-gpib":
            self.backend.clear(self._descr)
            self.backend.close(self._descr)
        elif self.lib == "prologix-gpib":
            # clear the instrument buffers
            self.backend.write(
                "clr", to_controller=True, address=self.instrument.address
            )

            # return instrument to local control
            self.backend.write("loc", to_controller=True)

            # unlink device from controller
            self.backend.devices.remove(self.instrument.address)

            # if no more devices are connected to the controller, close it
            if len(self.backend.devices) == 0:
                self.backend.close()
                GPIB.prologix_controllers.pop(self.prologix_address)

        self.connected = False

    def __repr__(self):
        return "GPIB"


class PrologixGPIBUSB:
    """
    Wraps serial communications with the Prologix GPIB-USB adapter unit.
    """

    @property
    def timeout(self):
        return self.serial_port.timeout

    @timeout.setter
    def timeout(self, timeout):
        self.serial_port.timeout = timeout

    def __init__(self, port):
        try:
            serial = importlib.import_module("serial")
        except ImportError:
            raise AdapterError(
                "Please install the PySerial library "
                "to connect a Prologix GPIB-USB adapter."
            )

        self.serial_port = serial.Serial(port=port, timeout=1)

        # set adapter to "controller" mode
        self.write("mode 1", to_controller=True)

        # instruments talk only when requested to
        self.write("auto 0", to_controller=True)

        # set timeout to 0.5 seconds
        self.write("read_tmo_ms 500", to_controller=True)

        # Do not append CR or LF to messages
        self.write("eos 3", to_controller=True)

        # Assert EOI with last byte to indicate end of data
        self.write("eoi 1", to_controller=True)

        # Append CR to responses from instruments to indicate message
        # termination
        self.write("eot_char 13", to_controller=True)
        self.write("eot_enable 1", to_controller=True)

        self.address = None
        self.devices = []

    def write(self, message, to_controller=False, address=None):
        if address and address != self.address:
            self.write(f"addr {address}", to_controller=True)
            self.address = address

            if address not in self.devices:
                self.devices.append(address)

        proper_message = message.encode() + b"\r"

        if to_controller:
            proper_message = b"++" + proper_message

        self.serial_port.write(proper_message)

        return "Success"

    def read(self, from_controller=False, address=None):
        if address and address != self.address:
            self.write(f"addr {address}", to_controller=True)
            self.address = address

            if address not in self.devices:
                self.devices.append(address)

        if not from_controller:
            self.write(f"read eoi", to_controller=True)

        response = self.serial_port.read_until(b"\r").decode().strip()

        return response

    def close(self):
        self.serial_port.close()


class PrologixGPIBLAN:
    """
    Wraps serial communications with the Prologix GPIB-ETHERNET adapter unit.

    The IP address of the adapter unit can be found with Prologix's Netfinder
    utility and GPIB configuration can be modified with Prologix's GPIB
    Configurator.
    """

    @property
    def timeout(self):
        return self.socket.gettimeout()

    @timeout.setter
    def timeout(self, timeout):
        self.socket.settimeout(timeout)

    def __init__(self, ip_address):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.socket.connect((ip_address, 1234))

        self.socket.settimeout(1)

        # set adapter to "controller" mode
        self.write("mode 1", to_controller=True)

        # instruments talk only when requested to
        self.write("auto 0", to_controller=True)

        # set timeout to 0.5 seconds
        self.write("read_tmo_ms 500", to_controller=True)

        # Do not append CR or LF to messages
        self.write("eos 3", to_controller=True)

        # Assert EOI with last byte to indicate end of data
        self.write("eoi 1", to_controller=True)

        # Append CR to responses from instruments to indicate message
        # termination
        self.write("eot_char 13", to_controller=True)
        self.write("eot_enable 1", to_controller=True)

        self.devices = []
        self.address = None

    def write(self, message, to_controller=False, address=None):
        if address and address != self.address:
            self.write(f"addr {address}", to_controller=True)
            self.address = address

            if address not in self.devices:
                self.devices.append(address)

        if to_controller:
            message = "++" + message

        write_to_socket(self.socket, message)

        return "Success"

    def read(self, from_controller=False, address=None):
        if address and address != self.address:
            self.write(f"addr {address}", to_controller=True)
            self.address = address

            if address not in self.devices:
                self.devices.append(address)

        if not from_controller:
            self.write(f"read eoi", to_controller=True)

        return read_from_socket(self.socket)

    def close(self):
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()


class USB(Adapter):
    """
    Handles communications with pure USB instruments through PyVISA or USBTMC.
    """

    timeout = 0.5

    # Get USB library
    if importlib.util.find_spec("pyvisa"):
        lib = "pyvisa"
    elif importlib.util.find_spec("usbtmc"):
        lib = "usbtmc"
    else:
        lib = None

    no_lib_msg = "No USB library was found! " "Please install either PyVISA or USBTMC."

    def connect(self):
        serial_number = str(self.instrument.address)

        if self.lib == "pyvisa":
            visa = importlib.import_module("pyvisa")

            manager = visa.ResourceManager()

            for address in manager.list_resources():
                if serial_number in address:
                    self.backend = manager.open_resource(
                        address, open_timeout=self.timeout
                    )
                    self.backend.timeout = self.timeout

        elif self.lib == "usbtmc":
            usbtmc = importlib.import_module("usbtmc")

            self.backend = usbtmc.Instrument("USB::" + serial_number + "::INSTR")

        else:
            raise AdapterError(f"invalid library specification, {self.lib}")

        self.connected = True

    def _write(self, message):
        self.backend.write(message)
        return "Success"

    def _read(self):
        return self.backend.read()

    def _query(self, question):
        self._write(question)
        time.sleep(self.delay)
        return self._read()

    def disconnect(self):
        self.backend.close()

        self.connected = False

    def __repr__(self):
        return "USB"


class Socket(Adapter):
    """
    Handles communications between sockets using Python's built-in socket module
    """

    ip_address = None
    port = None

    read_termination = "\r"
    write_termination = "\r"
    timeout = 1

    kwargs = ["read_termination", "write_termination", "timeout"]

    def connect(self):
        if self.connected:
            self.disconnect()

        self.backend = socket.socket()

        self.backend.settimeout(self.timeout)

        address = self.instrument.address
        remote_ip_address, remote_port = address.split("::")

        self.backend.connect((remote_ip_address, int(remote_port)))

        self.connected = True

    def _write(self, message):
        write_to_socket(
            self.backend,
            message,
            termination=self.write_termination,
            timeout=self.timeout,
        )

        return "Success"

    def _read(self, **kwargs):
        termination = kwargs.pop("termination", self.read_termination)
        timeout = kwargs.pop("timeout", self.timeout)

        return read_from_socket(
            self.backend, termination=termination, timeout=timeout, **kwargs
        )

    def _query(self, question, **kwargs):
        self._write(question)
        return self._read(**kwargs)

    def disconnect(self):
        # Clear out any unread messages
        unread = self._read(decode=False, nbytes=np.inf)
        while unread:
            unread = self._read(decode=False, nbytes=np.inf)

        self.backend.shutdown(socket.SHUT_RDWR)
        self.backend.close()

        self.connected = False

    def __repr__(self):
        return "Socket"


class ModbusSerial(Adapter):
    """
    (TO BE REMOVED)

    Handles communications with modbus serial instruments through the
    Minimal ModbusSerial package
    """

    # Common defaults
    slave_mode = "rtu"
    baud_rate = 38400
    timeout = 0.05
    byte_size = 8
    stop_bits = 1
    parity = "N"
    delay = 0.05

    kwargs = ["close_port_after_each_call", "slave_mode", "byte_order"]

    # For traffic control of modbus adapters using the same serial ports
    adapters = {}

    _busy = False

    # Get Minimal Modbus library
    if importlib.util.find_spec("minimalmodbus"):
        lib = "minimalmodbus"
    else:
        lib = None

    no_lib_msg = (
        "No ModbusSerial library was found! " "Please install the minimalmodbus library"
    )

    @property
    def busy(self):
        return bool(
            sum([adapter._busy for adapter in ModbusSerial.adapters.get(self.port, [])])
        )

    @busy.setter
    def busy(self, busy):
        self._busy = busy

    def __repr__(self):
        return "ModbusSerial"

    def connect(self):
        minimal_modbus = importlib.import_module("minimalmodbus")

        # Get port and channel
        self.port, self.channel = self.instrument.address.split("::")

        if self.port in ModbusSerial.adapters:
            ModbusSerial.adapters[self.port].append(self)
        else:
            ModbusSerial.adapters[self.port] = [self]

        # Handshake with instrument
        self.backend = minimal_modbus.Instrument(
            self.port, int(self.channel), mode=self.slave_mode
        )

        self.backend.serial.baudrate = self.baud_rate
        self.backend.serial.timeout = self.timeout
        self.backend.serial.bytesize = self.byte_size
        self.backend.serial.parity = self.parity
        self.backend.serial.stopbits = self.stop_bits
        self.backend.close_port_after_each_call = True
        time.sleep(self.delay)

        self.connected = True

    def _write(self, register, message, dtype="uint16", byte_order=0):
        if dtype == "uint16":
            self.backend.write_register(register, message)
        elif dtype == "float":
            self.backend.write_float(register, message, byteorder=byte_order)
        time.sleep(self.delay)

        return "Success"

    def _read(self, register, dtype="uint16", byte_order=0):
        self.backend.serial.timeout = self.timeout

        if dtype == "uint16":
            return self.backend.read_register(register)
        elif dtype == "float":
            return self.backend.read_float(register, byteorder=byte_order)

    def disconnect(self):
        if not self.backend.close_port_after_each_call:
            self.backend.serial.close()
        self.connected = False

    @staticmethod
    def locate():
        return Serial.locate()


class Modbus(Adapter):
    """
    Handles communication with instruments via the Modbus communication
    protocol, over either TCP or serial ports, using PyModbus.
    """

    kwargs = (
        "slave id",
        "byte order",
        "word order",
        "baud rate",
        "timeout",
        "byte size",
        "stop bits",
        "parity",
        "delay",
    )

    slave_id = 0
    # Byte and word order is either little-endian (<) or big-endian (>)
    byte_order = ">"
    word_order = ">"

    types = [
        "8bit_uint",
        "16bit_uint",
        "32bit_uint",
        "64bit_uint",
        "8bit_int",
        "16bit_int",
        "32bit_int",
        "64bit_int",
        "16bit_float",
        "32bit_float",
        "64bit_float",
    ]

    baud_rate = 19200
    timeout = 0.05
    byte_size = 8
    stop_bits = 1
    parity = "N"
    delay = 0.05

    _protocol = None

    # This dict contains all active Modbus serial adapters. When a new adapter
    # is initialized with the same com port as an existing one, it uses the
    # same Modbus client object as its backend (they are differentiated by
    # their slave IDs).
    _serial_adapters = {}

    # Locate PyModbus library
    if importlib.util.find_spec("pymodbus"):
        lib = "pymodbus"
    else:
        lib = None

    @property
    def busy(self):
        if self._protocol == "Serial":
            return bool(
                sum([
                    adapter._busy for adapter in
                    Modbus._serial_adapters.get(self.port, [])
                ])
            )
        else:
            return self._busy

    @busy.setter
    def busy(self, busy):
        self._busy = busy

    def connect(self):
        client = importlib.import_module(".client", package="pymodbus")

        # Get port (Serial) or address & port (TCP)
        address = self.instrument.address.split("::")

        if re.match("\d+\.\d+\.\d+\.\d+", address[0]):
            # Modbus TCP
            self._protocol = "TCP"

            if len(address) == 1:
                address.append(502)  # standard Modbus TCP port

            self.backend = client.ModbusTcpClient(
                host=address[0],
                port=int(address[1])
            )

        else:
            # Modbus Serial
            self._protocol = "Serial"

            if len(address) == 1:

                # assume slave id is zero if not specified
                port, slave_id = address[0], 0

            else:
                port, slave_id = address

            if port in Modbus._serial_adapters:

                Modbus._serial_adapters[port].append(self)

                # use existing backend
                self.backend = Modbus._serial_adapters[port].backend

            else:

                Modbus._serial_adapters[port] = [self]

                self.backend = client.ModbusSerialClient(
                    address=port,
                    baudrate=self.baud_rate,
                    bytesize=self.byte_size,
                    parity=self.parity,
                    stopbits=self.stop_bits,
                )

            self.slave_id = slave_id

        # Get data reading/writing utility classes
        payload_module = importlib.import_module(".payload", package="pymodbus")

        # Utility for encoding data
        self._builder_cls = payload_module.BinaryPayloadBuilder

        # Utility for decoding data
        self._decoder_cls = payload_module.BinaryPayloadDecoder

        self.connected = True

    def _write(self, func_code, address, values, _type=None):
        """
        Write values to coils (func_code = 5 [single] or 15 [multiple]) or
        holding registers (func_code = 6 [single] or 16 [multiple]).

        The Modbus data type for decoding registers is specified by the `_type`
        argument. Valid values for `_type` are listed in the `_types` attribute.
        """

        values = np.array([values]).flatten()

        if func_code not in [5, 15, 6, 16]:
            raise ValueError(f"invalid Modbus function code {func_code}")

        if _type and _type not in self.types:
            raise TypeError(
                "invalid _type argument; must be one of:\n"
                + ", ".join(self.types)
            )

        if "5" in str(func_code):
            # Write coils

            bool_values = [bool(value) for value in values]

            response = self.backend.write_coils(
                address, bool_values, slave=self.slave_id
            )

            if response.function_code == 15:
                return "Success"
            else:
                raise AdapterError(
                    f"error writing to coil(s) on {self.instrument.name}"
                )

        else:
            # Write registers

            if _type is None:
                _type = "16bit_uint"

            builder = self._builder_cls(
                byteorder=self.byte_order, wordorder=self.word_order
            )

            for value in values:
                builder.__getattribute__("add_" + _type)(value)

            register_values = builder.to_registers()

            response = self.backend.write_registers(
                address, register_values, slave=self.slave_id
            )

            if response.function_code == 16:
                return "Success"
            else:
                raise AdapterError(
                    f"error writing to register(s) on {self.instrument.name} "
                    "for writing coils/registers"
                )

    def _read(self, func_code, address, count=1, _type=None):
        """
        Read from coils (func_code = 1), discrete inputs (func_code = 2),
        holding registers (func_code = 3), or input registers (func_code = 4).

        A single data unit is read by specifying the function code
        (`func_code`) and address; the data can be converted to the desired
        data type (`_type` = `int` or `float`). Multiple sequential addresses
        are read by specifying the count.
        """

        if _type and _type not in self.types:
            raise TypeError(
                "invalid _type argument; must be one of:\n" ", ".join(self.types)
            )

        # Enumerate modbus read functions
        read_functions = {
            1: self.backend.read_coils,
            2: self.backend.read_discrete_inputs,
            3: self.backend.read_holding_registers,
            4: self.backend.read_input_registers,
        }

        if func_code in [1, 2]:
            # Read coils or discrete inputs

            response = read_functions[func_code](
                address, count=count, slave=self.slave_id
            ).bits

            bits = [bool(bit) for bit in response][:count]

            if len(bits) == 1:
                return bits[0]
            else:
                return bits

        elif func_code in [3, 4]:
            # Read holding registers or input registers

            registers = read_functions[func_code](
                address, count=count, slave=self.slave_id
            ).registers

            decoder = self._decoder_cls.fromRegisters(
                registers, byteorder=self.byte_order, wordorder=self.word_order
            )

            n_values = int(16 * count / (int(_type.split("bit")[0])))

            values = [
                decoder.__getattribute__("decode_" + _type)() for _ in range(n_values)
            ]

            if len(values) == 1:
                return values[0]
            else:
                return values

        else:
            # Invalid function code
            raise ValueError(
                f"invalid Modbus function code {func_code} for "
                "reading coils/registers"
            )

    def _query(self, *args, **kwargs):
        """Alias of `_read` method"""
        return self._read(*args, **kwargs)

    def disconnect(self):
        self.backend.close()


class Phidget(Adapter):
    """
    Handles communications with Phidget devices
    """

    delay = 0.2
    timeout = 5

    if importlib.util.find_spec("Phidget22"):
        lib = "phidget"
    else:
        lib = None

    no_lib_msg = (
        "Phidget library was not found! " "Please install (pip[3] install Phidget22)."
    )

    def __repr__(self):
        return "Phidget"

    def connect(self):
        address_parts = self.instrument.address.split("::")
        address_parts = [int(part) for part in address_parts]

        serial_number = address_parts[0]

        self.PhidgetException = importlib.import_module(
            "Phidget22.PhidgetException"
        ).PhidgetException

        self.backend = self.instrument.device_class()

        self.backend.setDeviceSerialNumber(serial_number)

        if len(address_parts) == 2:
            self.backend.setChannel(address_parts[1])
        if len(address_parts) == 3:
            self.backend.setHubPort(address_parts[1])
            self.backend.setChannel(address_parts[2])

        self.backend.openWaitForAttachment(1000 * self.timeout)

        self.connected = True

    def _write(self, parameter, value):
        self.backend.__getattribute__("set" + parameter)(value)
        return "Success"

    def _query(self, parameter):
        return self.backend.__getattribute__("get" + parameter)()

    def disconnect(self):
        self.backend.close()
        self.connected = True


supported = {
    key: value
    for key, value in vars().items()
    if type(value) is type and issubclass(value, Adapter)
}

kwargs = []
for cls in supported.values():
    for kwarg in cls.kwargs:
        kwargs.append(kwarg)



import numbers, importlib
import numpy as np
from empyric.adapters import *
from empyric.collection.instrument import *
from empyric.types import Float, String, Integer, Array


class Keithley2110(Instrument):
    """
    Keithley 2110 digital multimeter instrument
    """

    name = "Keithley2110"

    supported_adapters = ((USB, {}),)

    knobs = ("voltage range", "current range")

    meters = ("voltage", "current", "temperature")

    _mode = None

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        if mode == "voltage":
            self.write('FUNC "VOLT"')
            self._mode = "voltage"
        if mode == "current":
            self.write('FUNC "CURR"')
            self._mode = "current"
        if mode == "temperature":
            self.write('FUNC "TCO"')
            self._mode = "temperature"

    @setter
    def set_voltage_range(self, voltage_range: Float):
        allowed_voltage_ranges = (0, 0.1, 1.0, 10.0, 100.0, 1000.0)

        if voltage_range in allowed_voltage_ranges[1:]:
            self.write("VOLT:RANG %.2e" % voltage_range)
        elif isinstance(voltage_range, numbers.Number):
            # Find nearest encapsulating voltage range
            try:
                nearest = np.argwhere(
                    voltage_range <= np.array(allowed_voltage_ranges)
                ).flatten()[0]
            except IndexError:
                nearest = -1

            self.set_voltage_range(allowed_voltage_ranges[nearest])

            Warning(
                "Given voltage range not an option, "
                f"setting to {allowed_voltage_ranges[nearest]} V instead"
            )

        elif voltage_range == 0.0:
            self.write("VOLT:RANG:AUTO")
        else:
            raise ValueError(f"voltage range choice {voltage_range} not permitted!")

    @setter
    def set_current_range(self, current_range: Float):
        allowed_current_ranges = (0.0, 0.01, 0.1, 1.0, 3.0, 10.0)

        if current_range in allowed_current_ranges:
            self.write("CURR:RANG %.2e" % current_range)
        elif isinstance(current_range, numbers.Number):
            # Find nearest encapsulating current range
            try:
                nearest = np.argwhere(
                    current_range <= np.array(allowed_current_ranges)
                ).flatten()[0]
            except IndexError:
                nearest = -1

            self.set_current_range(allowed_current_ranges[nearest])

            Warning(
                "Given current range not an option, "
                f"setting to {allowed_current_ranges[nearest]} A instead"
            )

        elif current_range == 0.0:
            self.write("CURR:RANG:AUTO")
        else:
            raise ValueError(f"current range choice {current_range} not permitted!")

    @measurer
    def measure_voltage(self) -> Float:
        if self.mode != "voltage":
            self.mode = "voltage"
        return float(self.query("READ?"))

    @measurer
    def measure_current(self) -> Float:
        if self.mode != "current":
            self.mode = "current"
        return float(self.query("READ?"))

    @measurer
    def measure_temperature(self) -> Float:
        if self.mode != "temperature":
            self.mode = "temperature"
        return float(self.query("READ?"))


class Keithley6500(Instrument):
    """
    Multimeter with 6.5 digits and high speed scanning and digitizing
    capabilities.

    For socket communication default port is 5025. If IP address is unknown,
    you can find or set it on the unit's Communication --> LAN menu.

    Uses TSP communication protocol.

    The `range` knob has separate values for the basic measurements (meter =
    'voltage' or 'current') and the digitized measurements (meter = 'fast
    voltages' or 'fast currents'). Be sure to set the desired range after
    switching meter types.
    """

    name = "Keithley6500"

    supported_adapters = (
        (Socket, {"write_termination": "\n", "read_termination": "\n", "timeout": 0.5}),
    )

    knobs = ("meter", "nplc", "range", "sample count", "sample rate", "trigger_source")

    meters = ("voltage", "current", "fast voltages", "fast currents")

    _trig_src = "trigger.EVENT_DISPLAY"
    meter = None

    @setter
    def set_meter(self, meter: String):
        if meter.lower() == "voltage":
            self.write(f"dmm.measure.func = dmm.FUNC_DC_VOLTAGE")
        elif meter.lower() == "current":
            self.write(f"dmm.measure.func = dmm.FUNC_DC_CURRENT")
        elif meter.lower() == "fast voltages":
            self.write(f"dmm.digitize.func = dmm.FUNC_DIGITIZE_VOLTAGE")
        elif meter.lower() == "fast currents":
            self.write(f"dmm.digitize.func = dmm.FUNC_DIGITIZE_CURRENT")
        else:
            raise print(f"Warning: invalid meter {meter} for Keithley6500")
            return self.get_meter()

        return meter.lower()

    @getter
    def get_meter(self) -> String:
        def validator(response):
            return re.match("dmm.FUNC", response)

        meter = self.query("print(dmm.measure.func)", validator=validator)

        if "dmm.FUNC_NONE" in meter:
            meter = self.query("print(dmm.digitize.func)")

            if "dmm.FUNC_NONE" in meter:
                raise ValueError(f"meter is undefined for {self.name}")

        meter_dict = {
            "dmm.FUNC_DC_CURRENT": "current",
            "dmm.FUNC_DC_VOLTAGE": "voltage",
            "dmm.FUNC_DIGITIZE_CURRENT": "fast currents",
            "dmm.FUNC_DIGITIZE_VOLTAGE": "fast voltages",
        }

        if meter in meter_dict:
            return meter_dict[meter]
        else:
            return ""

    @setter
    def set_sample_count(self, count: Integer):
        if "fast" in self.meter:  # pylint: disable=unsupported-membership-test
            self.write(f"dmm.digitize.count = {count}")
        else:
            return 1

    @getter
    def get_sample_count(self) -> Integer:
        if "fast" in self.meter:  # pylint: disable=unsupported-membership-test
            response = self.query("print(dmm.digitize.count)")

            if "nil" in response:
                return 0
            else:
                return int(response)
        else:
            return 1

    @setter
    def set_sample_rate(self, rate: Integer):
        if "fast" in self.meter:  # pylint: disable=unsupported-membership-test
            self.write(f"dmm.digitize.samplerate = {rate}")
        else:
            return 0

    @getter
    def get_sample_rate(self) -> Integer:
        if "fast" in self.meter:  # pylint: disable=unsupported-membership-test
            response = self.query("print(dmm.digitize.samplerate)")

            if "nil" in response:
                return 0
            else:
                return int(response)
        else:
            return 0

    @setter
    def set_nplc(self, nplc: Integer):
        if self.meter in ["voltage", "current"]:
            self.write(f"dmm.measure.nplc = {nplc}")
        else:
            return 0

    @getter
    def get_nplc(self) -> Integer:
        if self.meter in ["voltage", "current"]:
            return int(self.query("print(dmm.measure.nplc)"))
        else:
            return 0

    @setter
    def set_range(self, _range: Float):
        meter = self.get_meter()

        if meter in ["voltage", "current"]:
            if _range == 0.0:
                self.write("dmm.measure.autorange = dmm.ON")
            else:
                self.write(f"dmm.measure.range = {_range}")
        elif meter in ["fast voltages", "fast currents"]:
            self.write("dmm.digitize.range = %.3e" % _range)
        else:
            return self.get_range()

    @getter
    def get_range(self) -> Float:
        meter = self.get_meter()

        if meter in ["voltage", "current"]:
            if self.query("print(dmm.measure.autorange)") == "dmm.ON":
                return 0.0
            else:
                return float(self.query("print(dmm.measure.range)"))

        elif meter in ["fast voltages", "fast currents"]:
            return float(self.query("print(dmm.digitize.range)"))

    @setter
    def set_trigger_source(self, trigger_source: String):
        valid_sources = {
            "front panel": "trigger.EVENT_DISPLAY",
            "front": "trigger.EVENT_DISPLAY",
            "external in": "trigger.EVENT_EXTERNAL",
            "ext": "trigger.EVENT_EXTERNAL",
        }

        if trigger_source.lower() in valid_sources:
            self._trig_src = valid_sources[trigger_source.lower()]
        else:
            raise ValueError(f"invalid trigger source for {self.name}")

    @getter
    def get_trigger_source(self) -> String:
        trigger_sources = {
            "trigger.EVENT_DISPLAY": "front panel",
            "trigger.EVENT_EXTERNAL": "external in",
        }

        return trigger_sources[self._trig_src]

    @measurer
    def measure_current(self) -> Float:
        if self.meter != "current":
            self.set_meter("current")

        return recast(self.query("print(dmm.measure.read())"))

    @measurer
    def measure_voltage(self) -> Float:
        if self.meter != "voltage":
            self.set_meter("voltage")

        return recast(self.query("print(dmm.measure.read())"))

    def _execute_fast_measurements(self):
        if "fast" not in self.meter:  # pylint: disable=unsupported-membership-test
            return

        trigger_src = {
            "front panel": "trigger.EVENT_DISPLAY",
            "external in": "trigger.EVENT_EXTERNAL",
        }[self.get_trigger_source()]

        self.write(
            "trigger.model.setblock(1, trigger.BLOCK_BUFFER_CLEAR, "
            "defbuffer1)\n"
            "trigger.model.setblock(2, trigger.BLOCK_WAIT, "
            f"{trigger_src})\n"
            "trigger.model.setblock(3, trigger.BLOCK_DELAY_CONSTANT, 0)\n"
            "trigger.model.setblock(4, trigger.BLOCK_MEASURE_DIGITIZE, "
            "defbuffer1, dmm.digitize.count)\n"
            "trigger.model.initiate()\n"
        )

        running = True

        running_states = [
            "trigger.STATE_BUILDING",
            "trigger.STATE_RUNNING",
            "trigger.STATE_PAUSED",
            "trigger.STATE_WAITING",
        ]

        failed_states = [
            "trigger.STATE_ABORTING",
            "trigger.STATE_ABORTED",
            "trigger.STATE_FAILED",
        ]

        state = ""

        while running:
            self.write("state, state, block_num = trigger.model.state()")

            state = self.query("print(state)")

            running = state in running_states

            time.sleep(0.25)

        if state in failed_states:
            raise RuntimeError(f'fast measurement failed; trigger state is "{state}"')

        readings = self.query(
            "printbuffer(1, defbuffer1.n, defbuffer1.readings)", nbytes=np.inf
        )

        return np.array([np.float64(reading) for reading in readings.split(", ")])

    @measurer
    def measure_fast_voltages(self) -> Array:
        if self.meter != "fast voltages":
            self.set_meter("fast voltages")

        fast_voltages = self._execute_fast_measurements()

        return fast_voltages

    @measurer
    def measure_fast_currents(self) -> Array:
        if self.meter != "fast currents":
            self.set_meter("fast currents")

        fast_currents = self._execute_fast_measurements()

        return fast_currents


class LabJackU6(Instrument):
    """
    LabJack U6 Multi-function DAQ
    """

    name = "LabJackU6"

    supported_adapters = (
        # custom setup below until I can get serial or modbus comms to work
        (Adapter, {}),
    )

    knobs = (
        "DAC0 ",
        "DAC1",
    )

    meters = (
        "AIN0",
        "AIN1",
        "AIN2",
        "AIN3",
        "device temperature",
        "temperature 0",
        "temperature 1",
        "temperature 2",
        "temperature 3",
    )

    def __init__(self, *args, **kwargs):
        u6 = importlib.import_module("u6")
        self.backend = u6.U6()

        if len(args) == 0 and "address" not in kwargs:
            kwargs["address"] = str(self.backend.serialNumber)

        Instrument.__init__(self, *args, **kwargs)

    def write(self, register, value):
        self.backend.writeRegister(register, value)

    def read(self, register):
        return self.backend.readRegister(register)

    @setter
    def set_DAC0(self, value: Float):
        self.write(5000, value)

    @setter
    def set_DAC1(self, value: Float):
        self.write(5002, value)

    @getter
    def get_DAC0(self) -> Float:
        self.read(5000)

    @getter
    def get_DAC1(self) -> Float:
        self.read(5002)

    @measurer
    def measure_AIN0(self) -> Float:
        return self.read(0)

    @measurer
    def measure_AIN1(self) -> Float:
        return self.read(2)

    @measurer
    def measure_AIN2(self) -> Float:
        return self.read(4)

    @measurer
    def measure_AIN3(self) -> Float:
        return self.read(6)

    @measurer
    def measure_device_temperature(self) -> Float:
        return self.backend.getTemperature() - 273.15

    @measurer
    def measure_temperature_0(self) -> Float:
        return self.read(0) / 37e-6 + self.measure_device_temperature()

    @measurer
    def measure_temperature_1(self) -> Float:
        return self.read(2) / 37e-6 + self.measure_device_temperature()

    @measurer
    def measure_temperature_2(self) -> Float:
        return self.read(4) / 37e-6 + self.measure_device_temperature()

    @measurer
    def measure_temperature_3(self) -> Float:
        return self.read(6) / 37e-6 + self.measure_device_temperature()


class LabJackT7(Instrument):
    """
    LabJack T7/T7-Pro DAQ

    Only reading the default 14 inputs as voltages is currently supported, but
    this may eventually be expanded.
    """

    name = "LabJackT7"

    supported_adapters = ((Modbus, {}),)

    knobs = ("DIO0", "DIO1", "DIO2", "DIO3", "DIO4", "DIO5", "DIO6", "DIO7")

    meters = (
        "AIN0",
        "AIN1",
        "AIN2",
        "AIN3",
        "AIN4",
        "AIN5",
        "AIN6",
        "AIN7",
        "AIN8",
        "AIN9",
        "AIN10",
        "AIN11",
        "AIN12",
        "AIN13",
        "AIN all",
        "device temperature",
    )

    def _set_DION(self, n, value: Integer):
        self.write(16, 2000 + n, value, _type="16bit_uint")

    def _get_DION(self, n) -> Integer:
        return self.read(3, 2000 + n, count=1, _type="16bit_uint")

    def _measure_AIN(self, n) -> Float:
        return self.read(3, 2 * n, count=2, _type="32bit_float")

    @setter
    def set_DIO0(self, value: Integer):
        self._set_DION(0, value)

    @getter
    def get_DIO0(self) -> Integer:
        return self._get_DION(0)

    @setter
    def set_DIO1(self, value: Integer):
        self._set_DION(1, value)

    @getter
    def get_DIO1(self) -> Integer:
        return self._get_DION(1)

    @setter
    def set_DIO2(self, value: Integer):
        self._set_DION(2, value)

    @getter
    def get_DIO2(self) -> Integer:
        return self._get_DION(2)

    @setter
    def set_DIO3(self, value: Integer):
        self._set_DION(3, value)

    @getter
    def get_DIO3(self) -> Integer:
        return self._get_DION(3)

    @setter
    def set_DIO4(self, value: Integer):
        self._set_DION(4, value)

    @getter
    def get_DIO4(self) -> Integer:
        return self._get_DION(4)

    @setter
    def set_DIO5(self, value: Integer):
        self._set_DION(5, value)

    @getter
    def get_DIO5(self) -> Integer:
        return self._get_DION(5)

    @setter
    def set_DIO6(self, value: Integer):
        self._set_DION(6, value)

    @getter
    def get_DIO6(self) -> Integer:
        return self._get_DION(6)

    @setter
    def set_DIO7(self, value: Integer):
        self._set_DION(7, value)

    @getter
    def get_DIO7(self) -> Integer:
        return self._get_DION(7)

    @measurer
    def measure_AIN0(self) -> Float:
        return self._measure_AIN(0)

    @measurer
    def measure_AIN1(self) -> Float:
        return self._measure_AIN(1)

    @measurer
    def measure_AIN2(self) -> Float:
        return self._measure_AIN(2)

    @measurer
    def measure_AIN3(self) -> Float:
        return self._measure_AIN(3)

    @measurer
    def measure_AIN4(self) -> Float:
        return self._measure_AIN(4)

    @measurer
    def measure_AIN5(self) -> Float:
        return self._measure_AIN(5)

    @measurer
    def measure_AIN6(self) -> Float:
        return self._measure_AIN(6)

    @measurer
    def measure_AIN7(self) -> Float:
        return self._measure_AIN(7)

    @measurer
    def measure_AIN8(self) -> Float:
        return self._measure_AIN(8)

    @measurer
    def measure_AIN9(self) -> Float:
        return self._measure_AIN(9)

    @measurer
    def measure_AIN10(self) -> Float:
        return self._measure_AIN(10)

    @measurer
    def measure_AIN11(self) -> Float:
        return self._measure_AIN(11)

    @measurer
    def measure_AIN12(self) -> Float:
        return self._measure_AIN(12)

    @measurer
    def measure_AIN13(self) -> Float:
        return self._measure_AIN(13)

    @measurer
    def measure_AIN_all(self) -> Array:
        """Reads all 14 analog inputs in a single call"""
        return self.read(4, 0, count=2 * 14, _type="32bit_float")

    @measurer
    def measure_device_temperature(self) -> Float:
        """Device temperature in C"""
        return self.read(4, 60052, count=2, _type="32bit_float") - 273.15



import time
import numpy as np
import struct
from empyric.adapters import *
from empyric.collection.instrument import *
from empyric.types import Float, Integer, Array


class SRSRGA(Instrument):
    """
    SRS Residual Gas Analyzer, a quadrupole mass spectrometer with mass ranges
    from 100 to 300 amu
    """

    name = "SRS-RGA"

    supported_adapters = (
        (Serial, {"baud_rate": 28800, "timeout": 300, "read_termination": "\n\r"}),
    )

    knobs = (
        "initialize" "filament current",
        "mass",
        "masses",
        "mass range",
        "ppsf",  # partial pressure sensitivity factor
        "tpsf",  # total pressure sensitivity factor
    )

    presets = {"filament current": 1}

    postsets = {"filament current": 0}

    meters = {"filament current", "single", "spectrum", "total pressure"}

    def __init__(self, *args, **kwargs):
        Instrument.__init__(self, *args, **kwargs)

    @setter
    def set_filament_current(self, current: Float):
        # current is in mA

        if current >= 3.5:
            self.query("FL3.5")
        elif current <= 0:
            self.query("FL0")
        else:
            self.query("FL" + f"{np.round(current, 2)}")

        time.sleep(5)

    @measurer
    def measure_filament_current(self) -> Float:
        return float(self.query("FL?"))

    @setter
    def set_mass(self, mass: Integer):
        self.write("ML" + f"{mass}")

    @getter
    def get_mass(self) -> Integer:
        return int(self.query("ML?"))

    @setter
    def set_masses(self, masses: Array):
        initial_mass, final_mass = masses[0], masses[-1]

        self.write("MI" + f"{int(initial_mass)}")
        self.write("MF" + f"{int(final_mass)}")

        self.mass_range = [initial_mass, final_mass]

    @getter
    def get_masses(self) -> Array:
        initial_mass = int(self.query("MI?"))
        final_mass = int(self.query("MF?"))

        self.mass_range = [initial_mass, final_mass]

        return np.arange(initial_mass, final_mass + 1, dtype=np.int64)

    @setter
    def set_mass_range(self, mass_range: Array):
        initial_mass, final_mass = mass_range

        self.write("MI" + f"{int(initial_mass)}")
        self.write("MF" + f"{int(final_mass)}")

        self.masses = np.arange(initial_mass, final_mass + 1)

    @getter
    def get_mass_range(self) -> Array:
        initial_mass = int(self.query("MI?"))
        final_mass = int(self.query("MF?"))

        self.masses = np.arange(initial_mass, final_mass + 1)

        return [initial_mass, final_mass]

    @setter
    def set_ppsf(self, value: Float):
        self.write(f"SP{value}")

    @getter
    def get_ppsf(self) -> Float:
        return float(self.query("SP?"))

    @setter
    def set_tpsf(self, value: Float):
        self.write(f"ST{value}")

    @getter
    def get_tpsf(self) -> Float:
        return float(self.query("ST?"))

    @measurer
    def measure_spectrum(self) -> Array:
        response = self.query("HS1", num_bytes=4 * (len(self.masses) + 1), decode=False)
        return (
            np.array(struct.unpack("<" + "i" * (len(self.masses) + 1), response))[:-1]
            * 1.0e-16
            / self.ppsf
            * 1000
        )

    @measurer
    def measure_single(self) -> Float:
        response = self.query("MR" + f"{int(self.mass)}", num_bytes=4, decode=False)
        return struct.unpack("<i", response)[0] * 1.0e-16 / self.ppsf * 1000

    @measurer
    def measure_total_pressure(self) -> Float:
        response = self.query("TP?", num_bytes=4, decode=False)
        return struct.unpack("<i", response)[0] * 1.0e-16 / self.tpsf * 1000



import argparse

try:
    import pytest
except ImportError:
    pytest = None

from empyric.experiment import Manager


# List of testable features.
# Tests are invoked at the command line with `empyric --test <feature>`
testable_features = [
    "experiment",  # tests Variable, Experiment and Manager classes (default)
    "variable",  # tests Knob, Meter, Parameter and Expression
    # Adapters
    "serial",
    "gpib",
    "usb",
    "modbus",
    "phidget",
]


def execute():
    parser = argparse.ArgumentParser(description="Empyric Command Line Interface")

    parser.add_argument(
        "-r",
        "--runcard",
        type=str,
        required=False,
        default=None,
        help="path of experiment runcard to execute",
    )

    parser.add_argument(
        "-d",
        "--directory",
        type=str,
        default=None,
        help="directory to write data and plots to",
    )

    parser.add_argument(
        "-t", "--test", nargs="*", help="test empyric installation and components"
    )

    args = parser.parse_args()

    if args.test is not None:
        if pytest is None:
            raise ImportError("pytest is not installed")

        if len(args.test) == 0:  # test main classes in experiment.py
            pytest.main(
                ["-r", "A", "--pyargs", "empyric.tests", "-k", "test_experiment", "-q"]
            )

        elif all([feature in testable_features for feature in args.test]):
            # test specified features
            test_names = ["test_" + name for name in args.test]

            pytest.main(
                [
                    "-r",
                    "A",
                    "--pyargs",
                    "empyric.tests",
                    "-k",
                    " or ".join(test_names),
                    "-q",
                ]
            )

        else:
            raise NotImplementedError(
                "\n"
                + "\n".join(
                    [
                        f"Requested test of {feature} is not implemented"
                        for feature in args.test
                        if feature not in testable_features
                    ]
                )
            )
    else:
        manager = Manager(runcard=args.runcard)
        manager.run(directory=args.directory)



import time
import select
import socket
import numbers
import numpy as np


# Tools for time-keeping
def convert_time(time_value):
    """
    If time_value is a string, converts a time of the form "[number] [units]"
    (e.g. "3.5 hours") to the time in seconds.
    If time_value is a number, just returns the same number
    If time_value is an array, iterates through the array doing either of the
    previous two operations on every element.

    :param time_value: (str/float) time value, possibly including units such as
    "hours"
    :return: (int) time in seconds
    """

    if np.size(time_value) > 1:
        return [convert_time(t) for t in time_value]

    if isinstance(time_value, numbers.Number):
        return time_value
    elif isinstance(time_value, str):
        # times can be specified in the runcard with units, such as minutes,
        # hours or days, e.g. "6 hours"
        time_parts = time_value.split(" ")

        if len(time_parts) == 1:
            return float(time_parts[0])
        elif len(time_parts) == 2:
            value, unit = time_parts
            value = float(value)
            return (
                value
                * {
                    "seconds": 1,
                    "second": 1,
                    "minutes": 60,
                    "minute": 60,
                    "hours": 3600,
                    "hour": 3600,
                    "days": 86400,
                    "day": 86400,
                }[unit]
            )
        else:
            raise ValueError(f"Unrecognized time format for {time_value}!")


class Clock:
    """
    Clock for keeping time in an experiment; works like a standard stopwatch
    """

    def __init__(self):
        self.start_time = self.stop_time = time.time()  # initially stopped
        self.stoppage = 0  # total time during which the clock has been stopped

    def start(self):
        if self.stop_time:
            self.stoppage += time.time() - self.stop_time
            self.stop_time = False

    def stop(self):
        if not self.stop_time:
            self.stop_time = time.time()

    def reset(self):
        self.__init__()

    @property
    def time(self):
        if self.stop_time:
            elapsed_time = self.stop_time - self.start_time - self.stoppage
        else:
            elapsed_time = time.time() - self.start_time - self.stoppage

        return elapsed_time


def find_nearest(allowed, value, overestimate=False, underestimate=False):
    """
    Find the closest in a list of allowed values to a given value.

    In some cases it might be beneficial to overestimate (choose nearest higher
    value) or underestimate (choose the nearest lower value)
    """

    if overestimate:
        diffs = np.array([abs(np.ceil(value - _value)) for _value in allowed])
        nearest = np.argwhere(diffs == np.min(diffs)).flatten()
    elif underestimate:
        diffs = np.array([abs(np.floor(value - _value)) for _value in allowed])
        nearest = np.argwhere(diffs == np.min(diffs)).flatten()
    else:
        diffs = np.array([abs(value - _value) for _value in allowed])
        nearest = np.argwhere(diffs == np.min(diffs)).flatten()

    if len(nearest) > 0:
        return allowed[nearest[0]]


# Tools for handling sockets
def get_ip_address(remote_ip="8.8.8.8", remote_port=80):
    """
    Connect to a server to resolve IP address; defaults to Google's DNS server
    if `remote_ip` and `remote_port` are not specified
    """

    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as tst_sock:
        tst_sock.connect((remote_ip, remote_port))
        return tst_sock.getsockname()[0]


def autobind_socket(_socket):
    """
    Automatically find the appropriate IP address and a free port to bind the
    socket to
    """

    bound = False

    ip_address = get_ip_address()

    for port in range(6174, 10000):
        try:
            _socket.bind((ip_address, port))
            bound = True
            break
        except OSError:
            pass

    if not bound:
        raise IOError(f"unable to bind socket at {ip_address} to any port!")

    return ip_address, port


def read_from_socket(
    _socket, nbytes=None, termination="\r", timeout=1, decode=True, chunk_size=4096
):
    """
    Read from a socket, with some effort taken to get the whole message.

    :param _socket: (socket.Socket) socket to read from.
    :param nbytes: (int) number of bytes to read; defaults to infinite.
    :param termination: (str/bytes/callable) if str or bytes, expected message
                        termination character(s); if callable, a function that
                        takes a message as its sole argument and returns True if
                        the message indicates that it is terminated, and False
                        otherwise.
    :param timeout: (numbers.Number) communication timeout in seconds;
                    used for both the `select.select` and `_socket.recv`
                    functions.
    :param decode: (bool) whether to return decoded string (True) or raw bytes
                   message (False); defaults to True.
    :param chunk_size: (int) number of bytes to read on each call to recv method.
    """

    # Block until the socket is readable or until timeout
    if timeout:
        readable = _socket in select.select([_socket], [], [], timeout)[0]
    else:
        readable = _socket in select.select([_socket], [], [])[0]

    if not readable:
        return None

    default_timeout = _socket.gettimeout()  # save default timeout

    _socket.settimeout(timeout)

    if nbytes is None:
        nbytes = np.inf

        if termination is None:
            raise ValueError(
                "nbytes must be a non-negative integer if termination is None"
            )

    null_responses = 0
    max_nulls = 3

    if type(termination) == str:
        termination = termination.encode()

    def is_terminated(message):
        if isinstance(termination, bytes):
            return termination in message
        elif callable(termination):
            return termination(message)
        else:
            return False

    message = b""

    while len(message) < nbytes and null_responses < max_nulls:
        part = b""

        remaining_bytes = nbytes - len(message)

        try:
            if remaining_bytes < chunk_size:
                part = _socket.recv(remaining_bytes)
            else:
                part = _socket.recv(chunk_size)

        except ConnectionResetError as err:
            print(f"Warning: {err}")
            break
        except socket.timeout:
            pass

        if len(part) > 0:
            message = message + part

            if is_terminated(message):
                break

        else:
            null_responses += 1

    _socket.settimeout(default_timeout)

    if decode:
        return message.decode().strip()
    else:
        return message


def write_to_socket(_socket, message, termination="\r", timeout=1):
    """
    Write a message to a socket, with care taken to get the whole message
    transmitted.

    :param _socket: (socket.Socket) socket to write to.
    :param message: (str) message to send.
    :param termination: (str/bytes) expected message termination character(s).
    :param timeout: (numbers.Number) timeout for `select.select` call.
    """

    # Block until the socket is writeable or until timeout
    if timeout:
        writeable = _socket in select.select([], [_socket], [], timeout)[1]
    else:
        writeable = _socket in select.select([], [_socket], [])[1]

    if not writeable:
        return 0

    bytes_message = (message + termination).encode()
    msg_len = len(bytes_message)

    failures = 0
    max_failures = 3

    total_sent = 0

    while total_sent < msg_len and failures < max_failures:
        sent = _socket.send(bytes_message[total_sent:])

        if sent == 0:
            failures += 1

        total_sent = total_sent + sent

    if total_sent < msg_len:
        raise IOError(f"Socket connection to {_socket.getsockname()} is broken!")

    return total_sent



import importlib
import numbers
import socket
import queue
import threading
import asyncio
import time
import select
import functools

import numpy as np
import pandas as pd

from empyric.tools import (
    convert_time,
    autobind_socket,
    read_from_socket,
    write_to_socket,
    get_ip_address,
)
from empyric.types import (
    recast,
    Boolean,
    Integer,
    Float,
    Toggle,
    OFF,
    ON,
    Array,
    String,
)
from empyric.types import supported as supported_types
from empyric.variables import Variable


class Routine:
    """
    A Routine periodically updates a set of `knobs` based on a given state of an
    experiment. The knobs argument should be a dictionary of the form
    {..., name: variable, ...}.

    The optional `start` and `end` arguments indicate when the routine should
    start and end. The default values are 0 and infinity, respectively. When
    updating, the routine compares these values to the `Time` value of the
    given state.

    All other arguments are fixed values or string dictionary keys,
    corresponding to variables values of the controlling experiment.
    """

    assert_control = True

    def __init__(
        self, knobs: dict, enable: String = None, start=0.0, end=np.inf, **kwargs
    ):
        self.knobs = knobs

        for knob in self.knobs.values():
            knob._controller = None  # to control access to knob

        self.enable = enable
        self.start = convert_time(start)
        self.end = convert_time(end)
        self.prepped = False
        self.finished = False

        for key, value in kwargs.items():
            self.__setattr__(key.replace(" ", "_"), value)

    @staticmethod
    def enabler(update):
        """
        Checks the enabling variable, start and stop times for the
        routine. If the enable variable evaluates to `True` and the time is
        between the start and stop times, the update method is called.

        If the enabling variable does not evaluate to `True`, then the routine's
        control of the knobs is revoked and no other action is taken.

        Otherwise, if the time is before the start time, the routine's prep
        method is called. If the time is after the end time, the routine's
        cleanup method is called.
        """

        @functools.wraps(update)
        def wrapped_update(self, state):
            if self.enable is not None and not state[self.enable]:
                for name, knob in self.knobs.items():
                    if knob._controller == self:
                        knob._controller = None
                return

            elif state["Time"] < self.start:
                if not self.prepped:
                    self.prep(state)
                    self.prepped = True

                return

            elif state["Time"] >= self.end:
                if not self.finished:
                    self.finish(state)
                    self.finished = True

                    # Release knobs from control
                    for name, knob in self.knobs.items():
                        if knob._controller == self:
                            knob._controller = None

                return

            else:

                if not self.prepped:
                    self.prep(state)

                for name, knob in self.knobs.items():
                    if knob._controller and knob._controller != self:
                        # take no action if another routine has control
                        return
                    elif self.assert_control:
                        # assert control if needed
                        knob._controller = self

                update(self, state)

        return wrapped_update

    def update(self, state):
        """
        Updates the knobs controlled by the routine based on the given state.

        :param state: (dict/Series) state of the calling experiment or process
        in the form, {..., variable: value, ...}
        :return: None
        """

        pass

    def terminate(self):
        """
        Signals the routine to stop
        """

        pass

    def prep(self, state):
        """
        Does any needed preparation before the routine starts
        """

        pass

    def finish(self, state):
        """
        Makes any final actions after the routine ends
        """


class Set(Routine):
    """
    Sets `knobs` to the given values.

    The `values` argument should be either a single key/value for all knobs or
    a 1D array of keys/values of the same length as the `knobs` argument.
    """

    def __init__(self, knobs: dict, values, **kwargs):
        Routine.__init__(self, knobs, **kwargs)

        if len(self.knobs) > 1 and np.ndim(values) == 0:
            # One value for all knobs
            self.values = [values] * len(self.knobs)
        else:
            self.values = values

    @Routine.enabler
    def update(self, state):
        for knob, value in zip(self.knobs.values(), self.values):
            if isinstance(value, str):
                value = state[value]

            if value:
                knob.value = value


class Ramp(Routine):
    """
    Ramps the set of `knobs` to given `target` values at `given rates`.

    The `target` or `rate` arguments can be single keys/values for all knobs, or
    1D arrays of keys/values.
    """

    def __init__(self, knobs: dict, targets, rates, **kwargs):
        """
        :param rates: (1D array) list of ramp rates
        """

        Routine.__init__(self, knobs, **kwargs)

        if np.ndim(rates) == 0:  # single rate for all knobs
            rates = [rates] * len(self.knobs)

        if np.ndim(targets) == 0:  # single target for all knobs
            targets = [targets] * len(self.knobs)

        self.rates = rates
        self.targets = targets

        self.now = None
        self.then = None

    @Routine.enabler
    def update(self, state):
        self.now = state["Time"]

        if self.then is None:
            self.then = state["Time"]

        for knob, rate, target in zip(self.knobs, self.rates, self.targets):
            if isinstance(target, String):
                target = state[target]

            if isinstance(rate, String):
                rate = state[rate]

            val_now = self.knobs[knob]._value

            if (
                not isinstance(target, numbers.Number)
                or not isinstance(rate, numbers.Number)
                or not isinstance(val_now, numbers.Number)
                or target == val_now
            ):
                # Do nothing if knob, rate or target are undefined
                continue

            sign = (target - val_now) / abs(target - val_now)

            val_nxt = val_now + sign * rate * (self.now - self.then)

            if sign * val_now <= sign * target < sign * val_nxt:
                self.knobs[knob].value = target
            else:
                self.knobs[knob].value = val_nxt

        self.then = state["Time"]


class Timecourse(Routine):
    """
    Ramps the `knobs` linearly through a series of `values` at given `times`.

    The `times` and `values` should be 1D or 2D arrays of values (or keys for
    `values`); if 2D then the first dimension needs to have the same length as
    the `knobs` argument.
    """

    def __init__(self, knobs: dict, times, values, **kwargs):
        """

        :param times: (1D/2D array) array or list of times relative to
                      the start time
        :param values: (1D/2D array) array or list of values
        :param kwargs: keyword arguments for Routine
        """

        Routine.__init__(self, knobs, **kwargs)

        if np.ndim(times) == 0:  # single value or file path
            self.times = [[times]] * len(self.knobs)
        elif np.ndim(times) == 1:  # single list of times for all knobs
            self.times = [times] * len(self.knobs)
        else:
            self.times = times

        if np.ndim(values) == 0:  # single value or file path
            self.values = [[values]] * len(self.knobs)
        elif np.ndim(values) == 1:  # single list of values for all knobs
            self.values = [values] * len(self.knobs)
        else:
            self.values = values

        # Times and/or values can be stored in CSV files
        for i, (t_elem, v_elem) in enumerate(zip(self.times, self.values)):
            if type(t_elem[0]) == str and ".csv" in t_elem[0]:
                df = pd.read_csv(t_elem[0])
                self.times[i] = df["times"].values

            if type(v_elem[0]) == str and ".csv" in v_elem[0]:
                df = pd.read_csv(v_elem[0])
                self.values[i] = [recast(val) for val in df["values"].values]

        self.times = np.array(convert_time(self.times)).astype(float)
        self.values = np.array(self.values, dtype=object)

        # Infer start and end times from times argument, if not given
        if "start" not in kwargs:
            self.start = np.min(self.times)

        if "end" not in kwargs:
            self.end = np.max(self.times)

    @Routine.enabler
    def update(self, state):
        knobs_times_values = zip(self.knobs, self.times, self.values)
        for knob, times, values in knobs_times_values:

            if np.min(times) > state["Time"] or np.max(times) < state["Time"]:
                continue

            j_last = np.argwhere(times <= state["Time"]).flatten()[-1]
            j_next = np.argwhere(times > state["Time"]).flatten()[0]

            last_time = times[j_last]
            next_time = times[j_next]

            last_value = values[j_last]
            next_value = values[j_next]

            if last_value in list(state.keys()):
                # if last_value is a variable name,
                # use that variable's value from state
                last_value = state[last_value]

            if next_value in list(state.keys()):
                next_value = state[next_value]

            # Ramp linearly between numerical values
            value = last_value + (next_value - last_value) * (
                state["Time"] - last_time
            ) / (next_time - last_time)

            self.knobs[knob].value = value

    def prep(self, state):

        # Validate values
        for knob, value_list in zip(self.knobs.keys(), self.values):
            for value in value_list:

                is_number = isinstance(value, numbers.Number)
                is_variable = value in list(state.keys())

                if not is_number and not is_variable:
                    raise ValueError(
                        f'value {value} given for knob {knob} in Timecourse '
                        f'routine is invalid; value must be a numeric type or '
                        f'the name of a variable in the updating state'
                    )

    def finish(self, state):
        # Upon routine completion, set each knob to its final value
        for knob, value in zip(self.knobs.values(), self.values[:, -1]):
            if knob._controller is None or knob._controller == self:

                if isinstance(value, String) and value in state:
                    knob.value = state[value]
                else:
                    knob.value = value


class Sequence(Routine):
    """
    Passes the `knobs` through a series of `values` regardless of time.

    The `values` should be a 1D or 2D array of keys/values; if 2D then the first
    dimension needs to have the same length as the `knobs` argument.
    """

    def __init__(self, knobs: dict, values, **kwargs):
        Routine.__init__(self, knobs, **kwargs)

        if np.ndim(values) == 0:  # single value or file path
            self.values = [[values]] * len(self.knobs)
        elif np.ndim(values) == 1:  # single list of values for all knobs
            self.values = [values] * len(self.knobs)
        else:
            self.values = values

        # Times and/or values can be stored in CSV files
        for i, v_elem in enumerate(self.values):
            if type(v_elem[0]) == str and ".csv" in v_elem[0]:
                df = pd.read_csv(v_elem[0])
                self.values[i] = [recast(val) for val in df["values"].values]

        self.values = np.array(self.values, dtype=object)

        self.iteration = 0

    @Routine.enabler
    def update(self, state):
        for knob, values in zip(self.knobs.values(), self.values):
            value = values[self.iteration]

            if isinstance(value, String):
                value = state[value]

            knob.value = value

        self.iteration = (self.iteration + 1) % len(self.values[0])

    def finish(self, state):
        # Upon routine completion, set each knob to its final value
        for knob, value in zip(self.knobs.values(), self.values[:, -1]):
            if knob._controller is None or knob._controller == self:

                if isinstance(value, String) and value in state:
                    knob.value = state[value]
                else:
                    knob.value = value


class Minimization(Routine):
    """
    Minimize a `meter`(/expression) influenced by the set of `knobs`, using
    simulated annealing.

    The `meter` argument is the expression or meter to be minimized. The
    `max_deltas` argument is an optional list/array of same length as `knobs`
    indicating the maximum change per step for each knob; if not specified,
    defaults to a list of ones. The `T0` and `T1` argumnets are the initial and
    final temperatures; if not specified, defaults to T0 = 0.0 and T1 = 0.0.
    The `samples` argument determines the number of meter values to average
    together for each step.
    """

    def __init__(
        self, knobs: dict, meter, max_deltas=None, T0=0.0, T1=0.0, samples=1, **kwargs
    ):
        Routine.__init__(self, knobs, **kwargs)

        self.meter = meter

        if max_deltas:
            self.max_deltas = np.array([max_deltas]).flatten()
        else:
            self.max_deltas = np.ones(len(self.knobs))

        self.T = T0
        self.T0 = T0
        self.T1 = T1
        self.samples = samples

        self.meter_values = []
        self.best_meter = None
        self.best_knobs = [None for _ in self.knobs]

        self.revert = False  # going back?

    @Routine.enabler
    def update(self, state):
        # Take no action if knobs values are undefined
        if None in [state[knob] for knob in self.knobs] or np.nan in [
            state[knob] for knob in self.knobs
        ]:
            return

        if not self.prepped:
            self.prep(state)
            self.prepped = True

        # Update temperature
        self.T = self.T0 + (self.T1 - self.T0) * (state["Time"] - self.start) / (
            self.end - self.start
        )

        # Get meter values
        meter_value = state[self.meter]

        # Check for valid new meter value
        if meter_value is not None and meter_value != np.nan:
            self.meter_values.append(meter_value)
        else:
            return

        # Check if enough samples have been measured
        if len(self.meter_values) <= self.samples:
            return

        # Check if found (or returned to) minimum
        if self.better(np.mean(self.meter_values)) or self.revert:
            # Record this new (or past) optimal state
            self.best_meter = np.mean(self.meter_values)
            self.best_knobs = [state[knob] for knob in self.knobs]

            # Generate and apply new knob settings
            new_knobs = self.best_knobs + self.max_deltas * (
                2 * np.random.rand(len(self.knobs)) - 1
            )

            for knob, new_value in zip(self.knobs.values(), new_knobs):
                knob.value = new_value

            self.meter_values = []
            self.revert = False

        else:
            for knob, best_val in zip(self.knobs.values(), self.best_knobs):
                knob.value = best_val

            self.meter_values = []
            self.revert = True

    def better(self, meter_value):
        if meter_value is None or meter_value == np.nan:
            return False

        if self.best_meter is None or self.best_meter == np.nan:
            return False

        change = meter_value - self.best_meter

        if self.T > 0:
            _rand = np.random.rand()
            return (change < 0) or (np.exp(-change / self.T) > _rand)
        else:
            return change < 0

    def prep(self, state):
        self.best_knobs = [state[knob] for knob in self.knobs]
        self.best_meter = state[self.meter]

    def finish(self, state):
        for knob, best_val in zip(self.knobs.values(), self.best_knobs):
            knob.value = best_val


class Maximization(Minimization):
    """
    Maximize a `meter`/expression influenced by the set of knobs;
    otherwise, works the same way as Minimize.
    """

    best_meter = -np.inf

    def better(self, meter_value):
        if meter_value is None or meter_value == np.nan:
            return False

        if self.best_meter is None or self.best_meter == np.nan:
            return False

        change = meter_value - self.best_meter

        if self.T > 0:
            _rand = np.random.rand()
            return (change > 0) or (np.exp(change / self.T) > _rand)
        else:
            return change > 0


class SocketServer(Routine):
    """
    Server routine for transmitting data to other experiments, local or remote,
    using the socket interface.

    Any knobs given in the `knobs` argument can be read and set by clients.
    Clients can also read the values of any variables of the controlling
    experiment, via the `state` argument of the `update` method.
    """

    assert_control = False

    def __init__(self, knobs: dict = None, **kwargs):
        if knobs is None:
            knobs = {}

        Routine.__init__(self, knobs, **kwargs)

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.ip_address, self.port = autobind_socket(self.socket)

        self.socket.listen(5)
        self.socket.settimeout(1)

        self.clients_queue = queue.Queue(1)
        self.clients_queue.put({})

        self.running = True

        self.state = {}

        self.acc_conn_thread = threading.Thread(target=self.accept_connections)

        self.acc_conn_thread.start()

        self.proc_requ_thread = threading.Thread(target=self.process_requests)

        self.proc_requ_thread.start()

    def terminate(self):
        # Kill client handling threads
        self.running = False
        self.acc_conn_thread.join()
        self.proc_requ_thread.join()

    def accept_connections(self):
        while self.running:
            try:
                (client, address) = self.socket.accept()

                clients = self.clients_queue.get()

                clients[address] = client

                self.clients_queue.put(clients)

                print(f"Client at {address} has connected")

            except socket.timeout:
                pass

            time.sleep(0.1)

        # Close sockets
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except OSError:  # socket was not connected
            pass
        self.socket.close()

        clients = self.clients_queue.get()

        for address, client in clients.items():
            try:
                client.shutdown(socket.SHUT_RDWR)
                client.close()
            except ConnectionError:
                pass

        # Return empty clients dict to queue so process_requests can exit
        self.clients_queue.put({})

    def process_requests(self):
        while self.running:
            clients = self.clients_queue.get()

            # Purge disconnected clients
            clients = {
                address: client
                for address, client in clients.items()
                if client is not None
            }

            for address, client in clients.items():
                outgoing_message = None

                try:
                    request = read_from_socket(client, chunk_size=1)
                except ConnectionError:
                    clients[address] = None
                    continue

                if client and request:
                    alias = " ".join(request.split(" ")[:-1])
                    value = request.split(" ")[-1]

                    if alias not in self.knobs and alias not in self.state:
                        outgoing_message = f"Error: invalid alias"

                    elif value == "settable?":
                        settable = (alias in self.knobs) and self.knobs[alias].settable

                        if settable:
                            outgoing_message = f"{alias} settable"
                        elif alias in self.state:
                            outgoing_message = f"{alias} read-only"
                        else:
                            outgoing_message = f"{alias} undefined"

                    elif value == "type?":
                        if alias in self.knobs:
                            _type = self.knobs[alias].type
                        elif alias in self.state:
                            _type = None
                            for supported_type in supported_types.values():
                                value = self.state[alias]
                                if isinstance(value, supported_type):
                                    _type = supported_type

                        else:
                            _type = None

                        outgoing_message = f"{alias} {_type}"

                    elif value == "?":  # Query of value
                        if alias in self.knobs:
                            _value = self.knobs[alias].value
                        elif alias in self.state:
                            _value = self.state[alias]
                        else:
                            _value = None

                        outgoing_message = f"{alias} {_value}"

                    else:  # Setting a value
                        knob_exists = alias in self.knobs

                        is_free = not getattr(self.knobs[alias], "_controller", None)

                        if knob_exists and is_free:
                            knob = self.knobs[alias]

                            knob.value = recast(value, to=knob.type)

                            outgoing_message = f"{alias} {knob.value}"

                        else:
                            outgoing_message = f"Error: cannot set {alias}"

                # Send outgoing message
                if outgoing_message is not None:
                    write_to_socket(client, outgoing_message)

                # Remove clients with problematic connections
                exceptional = client in select.select([], [], [client], 0)[2]

                if exceptional:
                    print(f"Client at {address} has a connection issue")
                    clients[address] = None

            self.clients_queue.put(clients)

            time.sleep(0.01)

    @Routine.enabler
    def update(self, state):
        self.state = state

    def __del__(self):
        if self.running:
            self.terminate()


class ModbusServer(Routine):
    """
    Server routine for transmitting data to other experiments, local or remote,
    using the Modbus over TCP/IP protocol.

    Any variables given in the `knobs` argument will be readable and writeable
    by connected clients. All variable values provided as the `state` argument
    of the `update` method can be read by clients.

    Because data is stored in statically assigned registers, only variables
    of boolean, toggle, integer or float types can be used. Variable values are
    stored in consecutive 4 registers (64 bits per value).

    Values of writeable variables (the `knobs` argument) will be stored in
    holding registers in the same order as given in the argument, starting from
    address 0. Values provided in the `state` argument of the `update` method
    will be stored in input registers in the same order as defined therein,
    starting from address 0. Each value in both sets of registers is stored as 5
    consecutive registers, 4 registers for the 64-bit value and 1 register for
    any metadata (i.e. data type). Note that the that `state` of an instance of
    `Experiment` has `Time` as its first entry.
    """

    assert_control = False

    def __init__(self, knobs: dict = None, meters=None, **kwargs):
        if knobs is None:
            knobs = {}

        Routine.__init__(self, knobs, **kwargs)

        self.meters = meters

        self.state = None  # set by update method

        # Check for PyModbus installation
        try:
            importlib.import_module("pymodbus")
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "pymodbus must be installed to run a Modbus server"
            )

        # Import tools from pymodbus
        datastore = importlib.import_module(".datastore", package="pymodbus")
        device = importlib.import_module(".device", package="pymodbus")
        payload = importlib.import_module(".payload", package="pymodbus")

        self._builder_cls = payload.BinaryPayloadBuilder
        self._decoder_cls = payload.BinaryPayloadDecoder

        DataBlock = datastore.ModbusSequentialDataBlock

        # Set up a PyModbus TCP Server
        datastore.ModbusSlaveContext.setValues = self.setValues_decorator(
            datastore.ModbusSlaveContext.setValues
        )

        self.slave = datastore.ModbusSlaveContext(
            ir=DataBlock.create(), hr=DataBlock.create()
        )

        self.context = datastore.ModbusServerContext(slaves=self.slave, single=True)

        self.identity = device.ModbusDeviceIdentification(
            info_name={
                "VendorName": "Empyric",
                "VendorUrl": "https://github.com/dmerthe/empyric",
                "ProductName": "Modbus Server",
                "ModelName": "Modbus Server",
            }
        )

        # Run server
        self.server = None  # assigned in _run_async_server
        self.ip_address = kwargs.get("address", get_ip_address())
        self.port = kwargs.get("port", 502)

        self.server_thread = threading.Thread(
            target=asyncio.run, args=(self._run_async_server(),)
        )

        self.server_thread.start()

    def setValues_decorator(self, setValues_method):
        """
        This decorator adds a `from_vars` kwarg to indicate that the intention
        is to update the registers from variable values. Otherwise, the wrapped
        method sets the corresponding variables instead.
        """

        @functools.wraps(setValues_method)
        def wrapped_method(self2, *args, from_vars=False):
            if from_vars:
                # update context from variables
                return setValues_method(self2, *args)
            else:
                # update variables according to the request

                fc_as_hex, address, values = args

                if fc_as_hex != 16:
                    print(
                        f"Warning: an attempt was made to write to a readonly "
                        f"register at address {address}"
                    )
                else:
                    if self.knobs:
                        variable = list(self.knobs.values())[address // 5]

                        controller = getattr(variable, "_controller", None)

                        if controller:
                            name = list(self.knobs.keys())[address // 5]

                            print(
                                f"Warning: an attempt was made to set {name}, "
                                "but it is currently controlled by "
                                f"{controller}."
                            )
                            return

                        decoder = self._decoder_cls.fromRegisters(values, byteorder=">")

                        if issubclass(variable._type, Boolean):
                            variable.value = decoder.decode_64bit_uint()
                        elif issubclass(variable._type, Toggle):
                            int_value = decoder.decode_64bit_uint()
                            variable.value = ON if int_value == 1 else OFF
                        elif issubclass(variable._type, Integer):
                            variable.value = decoder.decode_64bit_int()
                        elif issubclass(variable._type, Float):
                            variable.value = decoder.decode_64bit_float()

        return wrapped_method

    async def _update_registers(self):
        if self.state is None:  # do nothing if state is undefined
            await asyncio.sleep(0.1)
            asyncio.create_task(self._update_registers())
            return

        # Store readwrite variable values in holding registers (fc = 3)
        builder = self._builder_cls(byteorder=">")

        for i, (name, variable) in enumerate(self.knobs.items()):
            value = variable._value

            # encode the value into the 4 registers
            if value is None or variable.type is None:
                builder.add_64bit_float(float("nan"))
            elif issubclass(variable.type, Boolean):
                builder.add_64bit_uint(value)
            elif issubclass(variable.type, Toggle):
                builder.add_64bit_uint(int(value in Toggle.on_values))
            elif issubclass(variable.type, Integer):
                builder.add_64bit_int(value)
            elif issubclass(variable.type, Float):
                builder.add_64bit_float(value)
            else:
                raise ValueError(
                    f"unable to update modbus server registers from value "
                    f"{value} of variable {name} with data type "
                    f"{variable.type}"
                )

            # encode the meta data
            meta_reg_val = {
                Boolean: 0,
                Toggle: 1,
                Integer: 2,
                Float: 3,
                Array: 4,
                String: 5,
            }.get(variable.type, -1)

            builder.add_16bit_int(meta_reg_val)

        # from_vars kwarg added with setValues_decorator above
        self.slave.setValues(3, 0, builder.to_registers(), from_vars=True)

        # Store readonly variable values in input registers (fc = 4)
        builder.reset()

        if self.meters:
            selection = self.state[self.meters]
        else:
            selection = self.state

        for i, (name, value) in enumerate(selection.items()):
            _type = None
            for supported_type in supported_types.values():
                if isinstance(value, supported_type):
                    _type = supported_type

            # encode the value into the 4 registers
            if value is None or _type is None:
                builder.add_64bit_float(float("nan"))
            elif _type == Boolean:
                builder.add_64bit_uint(value)
            elif _type == Toggle:
                builder.add_64bit_uint(int(value in Toggle.on_values))
            elif _type == Integer:
                builder.add_64bit_int(value)
            elif _type == Float:
                builder.add_64bit_float(value)
            else:
                raise ValueError(
                    f"unable to update modbus server registers from value "
                    f"{value} of variable {name} with data type "
                    f"{_type}"
                )

            # encode the meta data
            type_int = {Boolean: 0, Toggle: 1, Integer: 2, Float: 3}.get(_type, -1)

            builder.add_16bit_int(type_int)

        # from_vars kwarg added with setValues_decorator above
        self.slave.setValues(4, 0, builder.to_registers(), from_vars=True)

        await asyncio.sleep(0.1)

        asyncio.create_task(self._update_registers())

    async def _run_async_server(self):
        asyncio.create_task(self._update_registers())

        server = importlib.import_module(".server", package="pymodbus")

        self.server = server.ModbusTcpServer(
            self.context, identity=self.identity, address=(self.ip_address, self.port)
        )

        try:
            await self.server.serve_forever()
        except asyncio.exceptions.CancelledError:
            # Server shutdown cancels the _update_registers task
            pass

    @Routine.enabler
    def update(self, state):
        self.state = state

    def terminate(self):
        asyncio.run(self.server.shutdown())


supported = {
    key: value
    for key, value in vars().items()
    if type(value) is type and issubclass(value, Routine)
}



import os
import datetime
import re
from typing import Union

import numpy as np
import pandas as pd

from empyric.collection.instrument import Instrument, setter, getter, measurer
from empyric.adapters import GPIB
from empyric.types import Toggle, ON, OFF, String, Float, Array


class Keithley2400(Instrument):
    """
    Keithley 2400 Sourcemeter, a 20 W power supply and picoammeter
    """

    name = "Keithley2400"

    supported_adapters = ((GPIB, {"delay": 0.1, "timeout": 0.5}),)

    # Available knobs
    knobs = (
        "voltage",
        "fast voltages",
        "current",
        "voltage range",
        "voltage limit",
        "current range",
        "current limit",
        "nplc",
        "delay",
        "output",
        "source",
        "meter",
        "remote sense",
        "source delay",
    )

    presets = {
        "source": "voltage",
        "meter": "current",
        "voltage": 0,
        "output": "ON",
        "source_delay": 0,
    }

    postsets = {"voltage": 0, "output": "OFF"}

    # Available meters
    meters = ("voltage", "current", "fast currents")

    fast_voltages = []

    #: allowed current range settings; zero indicate auto-range
    current_ranges = (0.0, 1.0e-6, 10.0e-6, 100.0e-6, 1.0e-3, 10.0e-3, 100.0e-3, 1.0)

    #: allowed voltage range settings; zero indicates auto-range
    voltage_ranges = (0.0, 0.2, 2.0, 20.0, 200.0)

    #: over-voltage protection level settings
    ovp_levels = (20.0, 40.0, 60.0, 80.0, 100.0, 120.0, 140.0, 160.0, 210.0)

    @setter
    def set_source(self, variable: String):
        if variable not in ["voltage", "current"]:
            raise ValueError('Source must be either "current" or "voltage"')

        self.write(":SOUR:CLE:AUTO OFF")  # disable auto output-off

        self.write(":SENS:FUNC:CONC OFF")  # disable concurrent measurements

        # Disabling concurrent measurements sets instrument to measure voltage
        self.set_meter(self.meter)

        self.set_output("OFF")

        if variable == "voltage":
            self.write(":SOUR:FUNC VOLT")
            self.current = None

        if variable == "current":
            self.write(":SOUR:FUNC CURR")
            self.voltage = None

    @getter
    def get_source(self) -> String:
        if self.query("SOUR:FUNC?").strip() == "VOLT":
            return "voltage"
        else:
            return "current"

    @setter
    def set_meter(self, variable: String):
        if variable not in ["voltage", "current"]:
            raise ValueError('Source must be either "current" or "voltage"')

        if variable == "voltage":
            self.write(':SENS:FUNC "VOLT"')
            self.write(":FORM:ELEM VOLT")

        if variable == "current":
            self.write(':SENS:FUNC "CURR"')
            self.write(":FORM:ELEM CURR")

        self.write(":TRIG:COUN 1")

    @getter
    def get_meter(self) -> String:
        if self.query(":SENS:FUNC?").strip() == "VOLT:DC":
            return "voltage"
        else:
            return "current"

    @setter
    def set_remote_sense(self, state: Toggle):
        self.set_output(OFF)

        if state.on:
            self.write(":SYST:RSEN ON")
        else:
            self.write(":SYST:RSEN OFF")

    @getter
    def get_remote_sense(self) -> Toggle:
        state = Toggle(self.query(":SYST:RSEN?").strip())

        return ON if state.on else OFF

    @setter
    def set_output(self, output: Toggle):
        if output:
            self.write(":OUTP ON")
        elif not output:
            self.write(":OUTP OFF")

            if not int(self.query(":OUTP?").strip()) == 0:
                raise RuntimeWarning(f"unable to turn off output of {self.name}")

        else:
            raise ValueError(f"Output setting {output} not recognized!")

    @getter
    def get_output(self) -> Toggle:
        return Toggle(self.query("OUTP?").strip())

    @measurer
    def measure_voltage(self) -> Float:
        if self.meter != "voltage":
            self.set_meter("voltage")

        if not self.output:
            self.set_output(ON)

        def validator(response):
            match = re.match(".\d\.\d+E.\d\d", response)
            return bool(match)

        return float(self.query(":READ?", validator=validator))

    @measurer
    def measure_current(self) -> Float:
        if self.meter != "current":
            self.set_meter("current")

        if not self.output:
            self.set_output(ON)

        def validator(response):
            match = re.match(".\d\.\d+E.\d\d", response)
            return bool(match)

        return float(self.query(":READ?", validator=validator))

    @setter
    def set_voltage(self, voltage: Float):
        if self.source != "voltage":
            Warning(f"Switching source mode to voltage!")
            self.set_source("voltage")

        if not self.output:
            self.set_output(ON)

        self.write(":SOUR:VOLT:LEV %.2E" % voltage)

    @getter
    def get_voltage(self) -> Float:
        return float(self.query(":SOUR:VOLT:LEV?"))

    @setter
    def set_current(self, current: Float):
        if self.source != "current":
            Warning(f"Switching source mode to current!")
            self.set_source("current")

        if not self.output:
            self.set_output("ON")

        self.write(":SOUR:CURR:LEV %.2E" % current)

    @getter
    def get_current(self) -> Float:
        return float(self.query(":SOUR:CURR:LEV?").strip())

    @setter
    def set_voltage_range(self, voltage_range: Float):
        if voltage_range in self.voltage_ranges:
            if self.source == "voltage":
                self.write(":SOUR:VOLT:RANGE %.2E" % voltage_range)
            else:
                if voltage_range == 0.0:
                    self.write(":SENS:VOLT:RANGE: AUTO ON")
                else:
                    self.write(":SENS:VOLT:RANGE %.2E" % voltage_range)
        else:
            first_line = (
                f"Given voltage range {voltage_range} "
                f"is not a valid value for {self.name}\n"
            )
            second_line = f"Valid values are {self.voltage_ranges}"
            raise ValueError(first_line + second_line)

    @getter
    def get_voltage_range(self) -> Float:
        if not hasattr(self, "source"):
            self.get_source()

        if self.source == "voltage":
            if Toggle(self.query(":SOUR:VOLT:RANGE:AUTO?").strip()):
                return 0.0
            else:
                return float(self.query(":SOUR:VOLT:RANG?").strip())

        else:
            if Toggle(self.query(":SENS:VOLT:RANGE:AUTO?").strip()):
                return 0.0
            else:
                return float(self.query(":SENS:VOLT:RANG?").strip())

    @setter
    def set_voltage_limit(self, voltage_limit: Float):
        if not hasattr(self, "source"):
            self.get_source()

        if self.source == "voltage":
            if voltage_limit in self.ovp_levels:
                self.write(":SOUR:VOLT:PROT %d" % int(voltage_limit))
            else:
                first_line = (
                    f"{self.name} is sourcing voltage, but given "
                    f"voltage limit {voltage_limit} "
                    f"is not a valid value\n"
                )
                second_line = f"Valid values are {self.ovp_levels}"
                raise ValueError(first_line + second_line)
        else:
            self.write(":SENS:VOLT:PROT %.2E" % voltage_limit)

    @getter
    def get_voltage_limit(self) -> Float:
        if not hasattr(self, "source"):
            self.get_source()

        if self.source == "voltage":
            return float(self.query(":SOUR:VOLT:PROT?").strip())
        else:
            return float(self.query(":SENS:VOLT:PROT?").strip())

    @setter
    def set_current_range(self, current_range: Float):
        if current_range in self.current_ranges:
            if self.source == "current":
                self.write(":SOUR:CURR:RANGE %.2E" % current_range)
            else:
                if current_range == 0.0:
                    self.write(":SENS:CURR:RANGE AUTO")
                else:
                    self.write(":SENS:CURR:RANGE %.2E" % current_range)
        else:
            first_line = (
                f"Given current range {current_range} "
                f"is not a valid value for {self.name}\n"
            )
            second_line = f"Valid values are {self.current_ranges}"
            raise ValueError(first_line + second_line)

    @getter
    def get_current_range(self) -> Float:
        if not hasattr(self, "source"):
            self.get_source()

        if self.source == "current":
            if Toggle(self.query(":SOUR:CURR:RANGE:AUTO?").strip()):
                return 0.0
            else:
                return float(self.query(":SOUR:CURR:RANG?").strip())

        else:
            if Toggle(self.query(":SENS:CURR:RANGE:AUTO?").strip()):
                return 0.0
            else:
                return float(self.query(":SENS:CURR:RANG?").strip())

    @setter
    def set_current_limit(self, current_limit: Float):
        self.write(":SENS:CURR:PROT %.2E" % current_limit)

    @getter
    def get_current_limit(self) -> Float:
        if self.source == "current":
            return float(self.query(":SOUR:CURR:PROT?".strip()))
        else:
            return float(self.query(":SENS:CURR:PROT?").strip())

    @setter
    def set_nplc(self, nplc: Float):
        if self.meter == "current":
            self.write(":SENS:CURR:NPLC %.2E" % nplc)
        elif self.meter == "voltage":
            self.write(":SENS:VOLT:NPLC %.2E" % nplc)

    @getter
    def get_nplc(self) -> Float:
        if not hasattr(self, "meter"):
            self.get_meter()

        if self.meter == "current":
            return float(self.query(":SENS:CURR:NPLC?"))
        elif self.meter == "voltage":
            return float(self.query(":SENS:VOLT:NPLC?"))

    @setter
    def set_delay(self, delay: Float):
        self.adapter.delay = delay

    @getter
    def get_delay(self) -> Float:
        return self.adapter.delay

    @setter
    def set_fast_voltages(self, voltages: Union[Array, String]):
        # import fast voltages, if specified as a path
        if type(voltages) == str:
            is_csv = ".csv" in voltages.lower()
            is_file = os.path.isfile(voltages)

            if not is_csv or not is_file:
                raise ValueError(
                    f"invalid fast voltages path for {self.name}; "
                    "a 1D numerical array or valid path to CSV file must be "
                    "provided."
                )

            voltage_data = pd.read_csv(voltages)

            columns = voltage_data.columns

            # Look for matching column name
            named = np.intersect1d(
                ["Voltage", "Fast Voltage", "Voltages", "Fast Voltages"], columns
            )

            if named:
                voltages = voltage_data[named[0]].values
            else:
                # If no matching column name, take the first column
                voltages = voltage_data[columns[0]].values

        if np.ndim(voltages) == 1:
            return np.array(voltages, dtype=float)  # --> self.fast_voltages
        else:
            raise ValueError(f"invalid fast voltages: {voltages}")

    @measurer
    def measure_fast_currents(self) -> Array:

        normal_timeout = self.adapter.timeout
        self.adapter.timeout = None  # measurements can take a while

        list_length = len(self.fast_voltages)

        if list_length == 0:
            raise ValueError(f"fast voltages for {self.name} have not been set")
        elif list_length <= 100:
            sub_lists = []
        else:  # instrument can only store 100 voltages
            sub_lists = [
                self.fast_voltages[i * 100 : (i + 1) * 100]
                for i in range(list_length // 100)
            ]

        if list_length % 100 > 0:
            sub_lists.append(self.fast_voltages[-(list_length % 100) :])

        self.set_meter("current")

        current_list = []

        self.set_source("voltage")
        self.set_meter("current")
        self.set_output("ON")
        self.write(":SOUR:VOLT:MODE LIST")

        for voltage_list in sub_lists:
            voltage_str = ", ".join(["%.4E" % voltage for voltage in voltage_list])

            self.write(":SOUR:LIST:VOLT " + voltage_str)
            self.write(":TRIG:COUN %d" % len(voltage_list))

            raw_response = self.query(":READ?")

            if raw_response is not None:
                raw_response = raw_response.strip()

                current_list += [
                    float(current_str) for current_str in raw_response.split(",")
                ]

            else:
                # truncated data; nullify measurement
                current_list = [np.nan for _ in self.fast_voltages]
                break

        self.write(":SOUR:VOLT:MODE FIX")
        self.write(":TRIG:COUN 1")

        self.adapter.timeout = normal_timeout

        return np.array(current_list)

    @setter
    def set_source_delay(self, delay: Float):
        self.write(":SOUR:DEL %.4E" % delay)

    @getter
    def get_source_delay(self) -> Float:
        response = self.query(":SOUR:DEL?")

        if response is not None:
            return response.strip()


class Keithley2460(Instrument):
    """
    Keithley 2460 Sourcemeter, a 100 W power supply and picoammeter
    """

    name = "Keithley2460"

    supported_adapters = ((GPIB, {}),)

    # Available knobs
    knobs = (
        "voltage",
        "fast voltages",
        "current",
        "voltage range",
        "voltage limit",
        "current range",
        "current limit",
        "nplc",
        "delay",
        "output",
        "source",
        "meter",
        "remote sense",
        "source delay",
    )

    presets = {
        "source": "voltage",
        "meter": "current",
        "voltage": 0,
        "output": "ON",
        "nplc": 1,
        "source delay": 0,
        "remote sense": "OFF",
    }

    postsets = {"voltage": 0, "output": "OFF"}

    # Available meters
    meters = ("voltage", "current", "fast currents")

    fast_voltages = None

    current_ranges = (1e-6, 10e-6, 100e-6, 1e-3, 10e-3, 100e-3, 1, 4, 5, 7)
    voltage_ranges = (0.2, 2, 7, 10, 20, 100)
    ovp_levels = (2, 5, 10, 20, 40, 60, 80, 100, 120, 140, 160, 180)

    @setter
    def set_source(self, variable):
        if variable not in ["voltage", "current"]:
            raise ValueError('Source must be either "current" or "voltage"')

        self.set_output("OFF")

        if variable == "voltage":
            self.write("SOUR:FUNC VOLT")
            self.current = None

        if variable == "current":
            self.write("SOUR:FUNC CURR")
            self.voltage = None

    @setter
    def set_meter(self, variable):
        if variable == "voltage":
            self.write('SENS:FUNC "VOLT"')
            self.write("DISP:VOLT:DIG 5")
        elif variable == "current":
            self.write('SENS:FUNC "CURR"')
            self.write("DISP:CURR:DIG 5")
        else:
            raise ValueError('Source must be either "current" or "voltage"')

    @setter
    def set_output(self, output):
        if output in [0, "OFF", "off"]:
            self.write(":OUTP OFF")
        elif output in [1, "ON", "on"]:
            self.write(":OUTP ON")
        else:
            raise ValueError(f"Output setting {output} not recognized!")

    @measurer
    def measure_voltage(self):
        if self.meter != "voltage":
            self.set_meter("voltage")

        if self.output == "ON":
            return float(self.query("READ?").strip())
        else:
            return np.nan

    @measurer
    def measure_current(self):
        if self.meter != "current":
            self.set_meter("current")

        if self.output == "ON":
            return float(self.query("READ?").strip())
        else:
            return 0

    @setter
    def set_voltage(self, voltage):
        if self.source != "voltage":
            Warning(f"Switching sourcing mode to voltage!")
            self.set_source("voltage")

            # turn output on if shut off when the source mode is changed
            self.set_output("ON")

        self.write("SOUR:VOLT:LEV %.4E" % voltage)

    @setter
    def set_current(self, current):
        if self.source != "current":
            Warning(f"Switching sourcing mode to current!")
            self.set_source("current")

            # turn output on if shut off when the source mode is changed
            self.set_output("ON")

        self.write("SOUR:CURR:LEV %.4E" % current)

    @setter
    def set_voltage_range(self, voltage_range):
        if voltage_range in self.voltage_ranges:
            if self.source == "voltage":
                self.write(":SOUR:VOLT:RANGE %.2E" % voltage_range)
            else:
                if voltage_range == "AUTO":
                    self.write(":SENS:VOLT:RANGE:AUTO ON")
                else:
                    self.write(":SENS:VOLT:RANGE %.2E" % voltage_range)
        else:
            first_line = (
                f"Given voltage range {voltage_range} "
                f"is not a valid value for {self.name}\n"
            )
            second_line = f"Valid values are {self.voltage_ranges}"
            raise ValueError(first_line + second_line)

    @setter
    def set_voltage_limit(self, voltage_limit):
        if self.source == "voltage":
            if int(voltage_limit) in self.ovp_levels:
                self.write(":SOUR:VOLT:PROT PROT%d" % int(voltage_limit))
            else:
                first_line = (
                    f"{self.name} is sourcing voltage, but given "
                    f"voltage limit {voltage_limit} "
                    f"is not a valid value\n"
                )
                second_line = f"Valid values are {self.ovp_levels}"
                raise ValueError(first_line + second_line)
        else:
            self.write(":SOUR:CURR:VLIM %.2E" % voltage_limit)

    @setter
    def set_current_range(self, current_range):
        if current_range in self.current_ranges:
            if self.source == "current":
                self.write(":SOUR:CURR:RANGE %.2E" % current_range)
            else:
                if current_range == "AUTO":
                    self.write(":SENS:CURR:RANGE:AUTO ON")
                else:
                    self.write(":SENS:CURR:RANGE %.2E" % current_range)
        else:
            first_line = (
                f"Given current range {current_range} "
                f"is not a valid value for {self.name}\n"
            )
            second_line = f"Valid values are {self.current_ranges}"
            raise ValueError(first_line + second_line)

    @setter
    def set_current_limit(self, current_limit):
        self.write(":SOUR:VOLT:ILIM %.2E" % current_limit)

    @setter
    def set_nplc(self, nplc):
        if self.meter == "current":
            self.write("CURR:NPLC %.2E" % nplc)
        elif self.meter == "voltage":
            self.write("VOLT:NPLC %.2E" % nplc)

    @setter
    def set_delay(self, delay):
        self.adapter.delay = delay

    @setter
    def set_fast_voltages(self, voltages):
        Keithley2400.set_fast_voltages(self, voltages)

    @measurer
    def measure_fast_currents(self):
        try:
            self.fast_voltages
        except AttributeError:
            raise ValueError("Fast IV sweep voltages have not been set!")

        if len(self.fast_voltages) == 0:
            raise ValueError("Fast IV sweep voltages have not been set!")

        path = self.name + "-fast_iv_measurement.csv"

        list_length = len(self.fast_voltages)

        if list_length >= 100:
            sub_lists = [
                self.fast_voltages[ # pylint: disable=unsubscriptable-object
                    i * 100 : (i + 1) * 100
                ]
                for i in range(list_length // 100)
            ]
        else:
            sub_lists = []

        if list_length % 100 > 0:
            sub_lists.append(
                self.fast_voltages[-(list_length % 100) :] # pylint: disable=unsubscriptable-object
            )

        current_list = []

        normal_timeout = self.adapter.timeout
        self.adapter.timeout = None  # the response times can be long

        start = datetime.datetime.now()
        for voltage_list in sub_lists:
            voltage_str = ", ".join(["%.4E" % voltage for voltage in voltage_list])
            self.write("SOUR:LIST:VOLT " + voltage_str)
            self.write("SOUR:SWE:VOLT:LIST 1, %.2e" % self.source_delay)
            self.write("INIT")
            self.write("*WAI")
            raw_response = self.query(
                'TRAC:DATA? 1, %d, "defbuffer1", SOUR, READ' % len(voltage_list)
            ).strip()

            current_list += [
                float(current_str) for current_str in raw_response.split(",")[1::2]
            ]

        self.adapter.timeout = normal_timeout  # put it back
        end = datetime.datetime.now()

        return np.array(current_list)

    @setter
    def set_source_delay(self, delay):
        if self.source == "voltage":
            self.write("SOUR:VOLT:DEL %.4e" % delay)
        else:
            self.write("SOUR:CURR:DEL %.4e" % delay)

    @setter
    def set_remote_sense(self, state):
        if bool(state) or state in ["ON", "1"]:
            self.write("VOLT:RSEN ON")
        else:
            self.write("VOLT:RSEN OFF")

        self.set_output("ON")

    @getter
    def get_remote_sense(self):
        if int(self.query("VOLT:RSEN?").strip()):
            return "ON"
        else:
            return "OFF"


class Keithley2651A(Instrument):
    """
    Keithley 2651A High Power (200 W) Sourcemeter
    """

    name = "Keithley2651A"

    supported_adapters = ((GPIB, {}),)

    # Available knobs
    knobs = (
        "voltage",
        "fast voltages",
        "current",
        "voltage range",
        "voltage limit",
        "current range",
        "current limit",
        "nplc",
        "output",
        "source",
        "meter",
        "source delay",
    )

    presets = {
        "voltage range": 40,
        "current range": 5,
        "voltage": 0,
        "output": "ON",
        "nplc": 1,
        "source": "voltage",
        "meter": "current",
        "source_delay": 0,
    }

    postsets = {"voltage": 0, "output": "OFF"}

    # Available meters
    meters = ("voltage", "current", "fast currents")

    fast_voltages = None

    @setter
    def set_source(self, variable):
        if variable == "voltage":
            self.write("smua.source.func =  smua.OUTPUT_DCVOLTS")
        elif variable == "current":
            self.write("smua.source.func = smua.OUTPUT_DCAMPS")
        else:
            raise ValueError('source must be either "current" or "voltage"')

    @setter
    def set_meter(self, variable):
        self.write("display.screen = display.SMUA")

        if variable == "current":
            self.write("display.smua.measure.func = display.MEASURE_DCAMPS")

        if variable == "voltage":
            self.write("display.smua.measure.func = display.MEASURE_DCVOLTS")

        # This sourcemeter does not require specifying the meter
        # before taking a measurement

    @setter
    def set_output(self, output):
        if output in [0, "OFF", "off"]:
            self.write("smua.source.output = smua.OUTPUT_OFF")
        elif output in [1, "ON", "on"]:
            self.write("smua.source.output = smua.OUTPUT_ON")
        else:
            raise ValueError(f"Ouput setting {output} not recognized!")

    @measurer
    def measure_voltage(self):
        if self.meter != "voltage":
            self.set_meter("voltage")

        if self.output == "ON":
            return float(self.query("print(smua.measure.v())").strip())
        else:
            return np.nan

    @measurer
    def measure_current(self):
        if self.meter != "current":
            self.set_meter("current")

        if self.output == "ON":
            return float(self.query("print(smua.measure.i())").strip())
        else:
            return 0

    @measurer
    def set_voltage(self, voltage):
        if self.source != "voltage":
            Warning(f"Switching sourcing mode!")
            self.set_source("voltage")

            # turn output on if shut off when the source mode is changed
            self.set_output("ON")

        self.write(f"smua.source.levelv = {voltage}")

    @setter
    def set_current(self, current):
        if self.source != "current":
            Warning(f"Switching sourcing mode!")
            self.set_source("current")

            # turn output on if shut off when the source mode is changed
            self.set_output("ON")

        self.write(f"smua.source.leveli = {current}")

    @setter
    def set_voltage_range(self, voltage_range):
        if voltage_range == "auto":
            self.write("smua.source.autorangev = smua.AUTORANGE_ON")
        else:
            self.write(f"smua.source.rangev = {voltage_range}")

    @setter
    def set_voltage_limit(self, voltage_limit):
        self.write(f"smua.source.limitv = {voltage_limit}")

    @setter
    def set_current_range(self, current_range):
        if current_range == "auto":
            self.write("smua.source.autorangei = smua.AUTORANGE_ON")
        else:
            self.write(f"smua.source.rangei = {current_range}")

    @setter
    def set_current_limt(self, current_limit):
        self.write(f"smua.source.limiti = {current_limit}")

    @setter
    def set_nplc(self, nplc):
        self.write(f"smua.measure.nplc = {nplc}")

    @setter
    def set_fast_voltages(self, voltages):
        self.fast_voltages = voltages

        # import fast voltages, if specified as a path
        if type(self.fast_voltages) == str:  # can be specified as a path
            try:
                fast_voltage_data = pd.read_csv(self.fast_voltages)
            except FileNotFoundError:
                # probably in an experiment data directory; try going up a level
                working_subdir = os.getcwd()
                os.chdir("..")
                fast_voltage_data = pd.read_csv(self.fast_voltages)
                os.chdir(working_subdir)

            columns = fast_voltage_data.columns

            fast_voltages = fast_voltage_data[columns[0]]

            self.fast_voltages = fast_voltages.astype(float).values

    @measurer
    def measure_fast_currents(self):
        try:
            if len(self.fast_voltages) == 0:
                raise ValueError("Fast IV sweep voltages have not been set!")
        except AttributeError:
            raise ValueError("Fast IV sweep voltages have not been set!")

        voltage_lists = []
        current_list = []

        list_length = 100  # maximum number of voltages to sweep at a time

        for i in range(len(self.fast_voltages) // list_length):
            voltage_lists.append(
                self.fast_voltages[i * list_length : (i + 1) * list_length]
            )

        remainder = len(self.fast_voltages) % list_length
        if remainder:
            voltage_lists.append(self.fast_voltages[-remainder:])

        normal_timeout = self.backend.timeout
        self.backend.timeout = 60  # give it up to a minute to do sweep

        for voltage_list in voltage_lists:
            voltage_string = ", ".join([f"{voltage}" for voltage in voltage_list])

            self.write("vlist = {%s}" % voltage_string)
            self.write(f"SweepVListMeasureI(smua, vlist, 0.01, {len(voltage_list)})")
            raw_response = self.query(
                f"printbuffer(1, {len(voltage_list)}, smua.nvbuffer1)"
            ).strip()
            current_list += [
                float(current_str) for current_str in raw_response.split(",")
            ]

            # hold last voltage until next sub-sweep
            self.set_voltage(voltage_list[-1])

        self.connection.timeout = normal_timeout  # put it back

        self.write("display.screen = display.SMUA")
        self.write("display.smua.measure.func = display.MEASURE_DCAMPS")

        return np.array(current_list)

    @setter
    def set_source_delay(self, delay):
        self.write(f"smua.source.delay = {delay}")



import re
from empyric.adapters import *
from empyric.collection.instrument import *
from empyric.types import Toggle, ON, OFF, String, Boolean, Integer, Float


class BRAX3000(Instrument):
    """
    BRAX 3000 series pressure gauge controller and meter
    """

    name = "BRAX3000"

    supported_adapters = (
        (Serial, {"baud_rate": 19200, "read_termination": "\r", "timeout": 1}),
    )

    knobs = (
        "ig state",
        "filament",
    )

    presets = {"filament": 1, "ig_state": "ON"}

    meters = (
        "cg1 pressure",
        "cg2 pressure",
        "ig pressure",
    )

    @setter
    def set_ig_state(self, state: Toggle):
        number = self.filament

        if state == ON:
            self.write(f"#IG{number} ON<CR>")
        if state == OFF:
            self.write(f"#IG{number} OFF<CR>")

        self.read()  # discard the response

    @getter
    def get_ig_state(self) -> Toggle:
        response = self.query("#IGS<CR>")

        if "ON" in response:
            return ON
        elif "OFF" in response:
            return OFF

    @setter
    def set_filament(self, number: Integer):
        pass

    @measurer
    def measure_cg1_pressure(self) -> Float:
        return float(self.query("#RDCG1<CR>")[4:-4])

    @measurer
    def measure_cg2_pressure(self) -> Float:
        return float(self.query("#RDCG2<CR>")[4:-4])

    @measurer
    def measure_ig_pressure(self) -> Float:
        def validator(response):
            match = re.search("\d\.\d+E-?\d\d", response)
            return bool(match)

        response = self.query("#RDIG<CR>", validator=validator)

        return float(re.findall("\d\.\d+E-?\d\d", response)[0])



import struct
from empyric.tools import find_nearest
from empyric.adapters import *
from empyric.collection.instrument import *
from empyric.types import Float, Array, Integer, String


class TekScope(Instrument):
    """
    Tektronix oscillscope of the TDS200, TDS1000/2000, TDS1000B/2000B,
    TPS2000 series

    NOT TESTED
    """

    name = "TekScope"

    supported_adapters = ((USB, {"timeout": 10}),)  # acquisitions can take a long time

    knobs = (
        "horz scale",
        "horz position",
        "ch1 scale",
        "ch1 position",
        "ch2 scale",
        "ch2 position",
        "ch3 scale",
        "ch3 position",
        "ch4 scale",
        "ch4 position",
        "trigger level",
    )

    meters = (
        "channel 1",
        "channel 2",
        "channel 3",
        "channel 4",
    )

    @setter
    def set_horz_scale(self, scale: Float):
        self.write("HOR:SCA %.3e" % scale)

    @setter
    def set_horz_position(self, position: Float):
        self.write("HOR:POS %.3e" % position)

    @setter
    def set_ch1_scale(self, scale: Float):
        self.write("CH1:SCA %.3e" % scale)

    @setter
    def set_ch2_scale(self, scale: Float):
        self.write("CH1:SCA %.3e" % scale)

    @setter
    def set_ch3_scale(self, scale: Float):
        self.write("CH1:SCA %.3e" % scale)

    @setter
    def set_ch4_scale(self, scale: Float):
        self.write("CH1:SCA %.3e" % scale)

    @setter
    def set_ch1_position(self, position: Float):
        self.write("CH1:POS %.3e" % position)

    @setter
    def set_ch2_position(self, position: Float):
        self.write("CH1:POS %.3e" % position)

    @setter
    def set_ch3_position(self, position: Float):
        self.write("CH1:POS %.3e" % position)

    @setter
    def set_ch4_position(self, position: Float):
        self.write("CH1:POS %.3e" % position)

    @setter
    def set_trigger_level(self, level: Float):
        self.write("TRIG:MAI:LEV %.3e" % level)

    def _measure_channel(self, channel):
        self.write("DAT:ENC ASCI")  # ensure ASCII encoding of data
        self.write("DAT:SOU CH%d" % channel)  # switch to channel 1

        scale_factor = float(self.query("WFMPRE:YMULT?"))
        zero = float(self.query("WFMPRE:YZERO?"))
        offset = float(self.query("WFMPRE:YOFF?"))

        self.write("ACQ:STATE RUN")  # acquire the waveform

        while int(self.query("BUSY?")):
            time.sleep(1)  # wait for acquisition to complete

        str_data = self.query("CURVE?").split(" ")[1].split(",")
        return np.array(
            [(float(datum) - offset) * scale_factor + zero for datum in str_data]
        )

    @measurer
    def measure_channel_1(self) -> Array:
        return self._measure_channel(1)

    @measurer
    def measure_channel_2(self) -> Array:
        return self._measure_channel(2)

    @measurer
    def measure_channel_3(self) -> Array:
        return self._measure_channel(3)

    @measurer
    def measure_channel_4(self) -> Array:
        return self._measure_channel(4)


class MulticompProScope(Instrument):
    """
    Multicomp Pro PC Oscilloscope.

    NOT TESTED
    """

    supported_adapters = (
        (USB, {"delay": 1}),
        # acquisitions can take a long time
    )
    """Supported adapters and options."""

    knobs = (
        "horz scale",
        "horz position",
        "scale ch1",
        "position ch1",
        "scale ch2",
        "position ch2",
        "sweep mode",
        "trigger level",
        "trigger source",
        "sweep mode",
        "resolution",
        "acquire",
        "state",
    )

    meters = (
        "channel 1",
        "channel 2",
    )

    presets = {
        "resolution": "100000",
        "sweep mode": "SINGLE",
        "trigger source": 1,
    }

    _time_scales = {
        2e-9: "2.0ns",
        5e-9: "5.0ns",
        10e-9: "10ns",
        20e-9: "20ns",
        50e-9: "50ns",
        100e-9: "100ns",
        200e-9: "200ns",
        500e-9: "500ns",
        1e-6: "1.0us",
        2e-6: "2.0us",
        5e-6: "5.0us",
        10e-6: "10us",
        20e-6: "20us",
        50e-6: "50us",
        100e-6: "100us",
        200e-6: "200us",
        500e-6: "500us",
        1e-3: "1.0ms",
        2e-3: "2.0ms",
        5e-3: "5.0ms",
        10e-3: "10ms",
        20e-3: "20ms",
        50e-3: "50ms",
        100e-3: "100ms",
        200e-3: "200ms",
        500e-3: "500ms",
        1: "1.0s",
        2: "2.0s",
        5: "5.0s",
        10: "10s",
        20: "20s",
        50: "50s",
        100: "100s",
    }

    _volt_scales = {
        2e-3: "2mv",
        5e-3: "5mv",
        10e-3: "10mv",
        20e-3: "20mv",
        50e-3: "50mv",
        100e-3: "100mv",
        200e-3: "200mv",
        500e-3: "500mv",
        1.0: "1v",
        2.0: "2v",
        5.0: "5v",
    }

    _volt_div_table = {0: 1e-3, 9: 1.0}

    _resolutions = {1e3: "1K", 1e4: "10K", 1e5: "100K", 1e6: "1M", 1e7: "10M"}

    _sweep_modes = ["AUTO", "NORMAL", "NORM", "SINGLE", "SING"]

    _acquiring = False

    @setter
    def set_horz_scale(self, scale):
        new_time_scale = find_nearest(list(self._time_scales.keys()), scale)

        self.write(":HORI:SCAL " + self._time_scales[new_time_scale])

    @getter
    def get_horz_scale(self):
        time_scales_rev = {val: key for key, val in self._time_scales.items()}

        time_scale_str = self.query(":HORI:SCAL?")[:-2]

        return time_scales_rev[time_scale_str]

    @setter
    def set_horz_position(self, offset):
        self.write(":HORI:OFFS " + str(offset))

    @getter
    def get_horz_position(self):
        return float(self.query(":HORI:OFFS?")[:-2])

    def _set_scale_ch(self, channel, scale):
        new_volt_scale = find_nearest(list(self._volt_scales.keys()), scale)

        scale_str = self._volt_scales[new_volt_scale]

        self.write((":CH%d:SCAL " % channel) + scale_str)

    def _get_scale_ch(self, channel):
        volt_scale_str = self.query(":CH%d:SCAL?" % channel)[:-2]

        if "mv" in volt_scale_str.lower():
            return 1e-3 * float(volt_scale_str[:-2])
        elif "v" in volt_scale_str.lower():
            return float(volt_scale_str[:-1])
        else:
            return float("nan")

    def _set_position_ch(self, channel, position):
        self.write((":CH%d:OFFS " % channel) + str(position))

    @setter
    def set_scale_ch(self, scale, channel=None):
        self._set_scale_ch(channel, scale)

    @getter
    def get_scale_ch(self, channel=None):
        return self._get_scale_ch(channel)

    @setter
    def set_position_ch(self, position, channel=None):
        self._set_position_ch(channel, position)

    @getter
    def get_position_ch(self, channel=None):
        return float(self.query(":CH%d:OFFS?" % channel)[:-2])

    @setter
    def set_sweep_mode(self, mode):
        if mode.upper() not in self._sweep_modes:
            raise ValueError(
                f"Invalid sweep mode! "
                f"Sweep mode must be one of: {', '.join(self._sweep_modes)}"
            )

        self.write(":TRIG:SING:SWE " + mode.upper())

    @getter
    def get_sweep_mode(self):
        return self.query(":TRIG:SING:SWE?")[:-2].upper()

    @setter
    def set_trigger_source(self, source):
        self.write(":TRIG:SING:EDGE:SOUR CH%d" % int(source))

    @getter
    def get_trigger_source(self):
        source = self.query(":TRIG:SING:EDGE:SOUR?")[:-2]

        if source == "CH1":
            return 1
        elif source == "CH2":
            return 2

    @setter
    def set_trigger_level(self, level_volts):
        trigger_source = self.get_trigger_source()

        scale = self._get_scale_ch(trigger_source)

        level_div = str(level_volts / scale)

        self.write(":TRIG:SING:EDGE:LEV " + level_div)

    @getter
    def get_trigger_level(self):
        trigger_source = self.get_trigger_source()

        scale = self._get_scale_ch(trigger_source)

        return float(self.query(":TRIG:SING:EDGE:LEV?")[:-2]) * scale

    @measurer
    def measure_trigger_status(self):
        return self.query(":TRIG:STATUS?")[:-2]

    @setter
    def set_resolution(self, resolution):
        if resolution not in self._resolutions:
            raise ValueError(
                f"Invalid memory depth! "
                f"Memory depth must be one of: {', '.join(self._resolutions)}"
            )

        self.write(":ACQ:DEPMEM " + self._resolutions[resolution])

    @getter
    def get_resolution(self):
        res_rev = {val: key for key, val in self._resolutions.items()}

        return res_rev[self.query(":ACQ:DEPMEM?")[:-2]]

    @setter
    def set_state(self, state):
        if state.upper() in ["RUN", "STOP"]:
            self.write(f":{state.upper()}")

    @setter
    def set_acquire(self, _):
        self._acquiring = True

        self.set_sweep_mode("SINGLE")

        trig_status = self.measure_trigger_status()

        if trig_status == "TRIG":
            # measurement has been triggered
            pass
        else:
            # sweep needs to be reset
            time.sleep(0.1)
            self.set_acquire(1)

    def _read_preamble(self, channel):
        info_dict = {}

        preamble = self.query(":WAV:PRE?", binary=True)

        header, info = preamble[:11], preamble[11:]

        if header != b"#9000000000":  # indicates empty data buffer
            try:
                # First 8 bytes is an integer signature, used for verification
                ver_int = 651058244139746640
                verified = struct.unpack("<q", info[:8])[0] == ver_int

                if verified:
                    if channel == 1:
                        scale = self._volt_div_table[
                            struct.unpack("<h", info[260:262])[0]
                        ]
                        zero = struct.unpack("<f", info[268:272])[0]
                    else:
                        scale = self._volt_div_table[
                            struct.unpack("<h", info[262:264])[0]
                        ]
                        zero = struct.unpack("<f", info[272:276])[0]

                    sample_rate = struct.unpack("<f", info[316:320])[0] * 1e6
                    # Hz

                    info_dict["scale"] = scale
                    info_dict["zero"] = zero
                    info_dict["sample rate"] = sample_rate
                else:
                    print(
                        f"Warning: {self.name} received unverified "
                        f"signature when reading from channel {channel}"
                    )

            except struct.error:
                print(
                    f"Warning: {self.name} received truncated preamble when "
                    f"reading from channel {channel}"
                )

        self._sample_rate = info["sample rate"]

        return info_dict

    def _read_data(self, channel):
        info = self._read_preamble(channel)

        if info:
            scale = info["scale"]
            zero = info["zero"]
        else:
            return None

        # Only 256k data points can be read at a time
        resolution = int(
            self.get_resolution().replace("K", "000").replace("M", "000000")
        )

        if resolution < 256000:  # can get all data in one pass
            self.write(":WAV:RANG 0, %d" % resolution)

            raw_response = self.query(":WAV:FETC?", binary=True)

            header, byte_data = raw_response[:11], raw_response[11:]

            data_len = int(header[2:]) // 2

            voltages = scale * (
                np.array(struct.unpack(f"<{data_len}h", byte_data), dtype=float) / 6400
                - zero
            )

        else:  # need to read data in chunks; max chunk size is 256k
            offset, size = 0, 200000

            voltages = np.empty(resolution)

            while offset + size <= resolution:
                self.write(":WAV:RANG %d, %d" % (offset, size))

                raw_response = self.query(":WAV:FETC?", binary=True)

                header, byte_data = raw_response[:11], raw_response[11:]

                data_len = int(header[2:]) // 2

                voltages[offset : offset + size] = scale * (
                    np.array(struct.unpack(f"<{data_len}h", byte_data), dtype=float)
                    / 6400
                    - zero
                )

                offset += size

        if len(voltages) != resolution:
            print(
                f"Warning: {self.name} received truncated data "
                f"for channel {channel}; discarding"
            )
            return None

        return voltages

    @measurer
    def measure_channel_1(self):
        if not self._acquiring:
            self.set_acquire(True)

        self._acquiring = False

        self.write(":WAV:BEG CH1")

        data = self._read_data(1)

        self.write(":WAV:END CH1")

        return data[1]

    @measurer
    def measure_channel_2(self):
        if not self._acquiring:
            self.set_acquire(True)

        self._acquiring = False

        self.write(":WAV:BEG CH2")

        data = self._read_data(2)

        self.write(":WAV:END CH2")

        return data[1]


class SiglentSDS1000(Instrument):
    """
    Siglent SDS1000 series digital oscilloscope
    """

    name = "SiglentSDS1000"

    supported_adapters = (
        (Socket, {"write_termination": "\n", "read_termination": "\n"}),
    )

    knobs = (
        "horz scale",
        "horz position",
        "ch1 scale",
        "ch1 position",
        "ch2 scale",
        "ch2 position",
        "ch3 scale",
        "ch3 position",
        "ch4 scale",
        "ch4 position",
        "trigger source",
        "trigger level",
        "acquire mode",
        "memory size",
        "averages",
    )

    meters = ("ch1 waveform", "ch2 waveform", "ch3 waveform", "ch4 waveform")

    horz_scales = [
        1e-9,
        2e-9,
        5e-9,
        10e-9,
        20e-9,
        50e-9,
        100e-9,
        200e-9,
        500e-9,
        1e-6,
        2e-6,
        5e-6,
        10e-6,
        20e-6,
        50e-6,
        100e-6,
        200e-6,
        500e-6,
        1e-3,
        2e-3,
        5e-3,
        10e-3,
        20e-3,
        50e-3,
        100e-3,
        200e-3,
        500e-3,
        1.0,
        2.0,
        5.0,
        10.0,
        20.0,
        50.0,
        100.0,  # seconds
    ]

    _trg_src = None

    # Time axis control
    @setter
    def set_horz_scale(self, scale: Float):
        nearest_scale = find_nearest(self.horz_scales, scale, overestimate=True)

        self.write(f"TDIV {nearest_scale}S")

        return nearest_scale

    @getter
    def get_horz_scale(self) -> Float:
        return float(self.query("TDIV?").split("TDIV ")[-1][:-1])

    @setter
    def set_horz_position(self, position: Float):
        self.write(f"TRDL {position}S")

    @getter
    def get_horz_position(self) -> Float:
        return float(self.query("TRDL?").split("TRDL ")[-1][:-1])

    # Channel control base functions
    def _set_chn_scale(self, n, scale):
        self.write("C%d:VDIV %.3eV" % (n, float(scale)))

    def _get_chn_scale(self, n):
        response = self.query("C%d:VDIV?" % n).split("C%d:VDIV " % n)[-1][:-1]

        return float(response)

    def _set_chn_position(self, n, position):
        self.write("C%d:OFST %.3eV" % (n, float(position)))

    def _get_chn_position(self, n):
        response = self.query("C%d:OFST?" % n).split("C%d:OFST " % n)[-1][:-1]

        return float(response)

    # Channel 1 control
    @setter
    def set_ch1_scale(self, scale: Float):
        self._set_chn_scale(1, scale)

    @getter
    def get_ch1_scale(self) -> Float:
        return self._get_chn_scale(1)

    @setter
    def set_ch1_position(self, position: Float):
        self._set_chn_position(1, position)

    @getter
    def get_ch1_position(self) -> Float:
        return self._get_chn_position(1)

    # Channel 2 control
    @setter
    def set_ch2_scale(self, scale: Float):
        self._set_chn_scale(2, scale)

    @getter
    def get_ch2_scale(self) -> Float:
        return self._get_chn_scale(2)

    @setter
    def set_ch2_position(self, position: Float):
        self._set_chn_position(2, position)

    @getter
    def get_ch2_position(self) -> Float:
        return self._get_chn_position(2)

    # Channel 3 control
    @setter
    def set_ch3_scale(self, scale: Float):
        self._set_chn_scale(3, scale)

    @getter
    def get_ch3_scale(self) -> Float:
        return self._get_chn_scale(3)

    @setter
    def set_ch3_position(self, position: Float):
        self._set_chn_position(3, position)

    @getter
    def get_ch3_position(self) -> Float:
        return self._get_chn_position(3)

    # Channel 4 control
    @setter
    def set_ch4_scale(self, scale: Float):
        self._set_chn_scale(4, scale)

    @getter
    def get_ch4_scale(self) -> Float:
        return self._get_chn_scale(4)

    @setter
    def set_ch4_position(self, position: Float):
        self._set_chn_position(4, position)

    @getter
    def get_ch4_position(self) -> Float:
        return self._get_chn_position(4)

    # Trigger control
    @setter
    def set_trigger_level(self, level: Float):
        trg_src = self.get_trigger_source()

        self.write(f"C{trg_src}:TRLV {level}V")

    @getter
    def get_trigger_level(self) -> Float:
        trg_src = self.get_trigger_source()

        return float(self.query(f"C{trg_src}:TRLV?").split("TRLV ")[-1][:-1])

    @setter
    def set_trigger_source(self, source: Integer):
        self.write(f"TRSE EDGE,SR,C{int(source)},OFF")

    @getter
    def get_trigger_source(self) -> Integer:
        return int(self.query("TRSE?").split("SR,C")[-1][0])

    @setter
    def set_acquire_mode(self, mode: String):
        if mode.upper() in ["SAMPLING", "PEAK_DETECT", "HIGH_RES"]:
            self.write(f"ACQW {mode.upper()}")
        elif mode.upper() == "AVERAGE":
            if not hasattr(self, "averages"):
                self.set_averages(4)

            self.write(f"ACQW AVERAGE,{int(self.averages)}")

        else:
            return None

    @getter
    def get_acquire_mode(self) -> String:
        response = self.query("ACQW?").split("ACQW ")[-1]

        if "AVERAGE" in response:
            return "AVERAGE"
        else:
            return response

    @setter
    def set_memory_size(self, size: Integer):
        self.write(f"MSIZ {int(size)}")

        return self.get_memory_size()

    @getter
    def get_memory_size(self) -> Integer:
        response = self.query("MSIZ?").split("MSIZ ")[-1]

        if "K" in response:
            response = float(response.replace("K", "e3"))
        elif "M" in response:
            response = float(response.replace("M", "e6"))

        return int(response)

    @setter
    def set_averages(self, averages: Integer):
        valid = [4] + [2**i for i in range(4, 11)]

        averages = find_nearest(valid, averages)

        self.write(f"AVGA {averages}")

    @getter
    def get_averages(self) -> Integer:
        return int(self.query("AVGA?").split("AVGA ")[-1])

    # Channel measurements
    def _measure_chn_waveform(self, n):
        # Set trigger mode to NORMAL
        if "NORM" not in self.query("TRMD?"):
            self.write("TRMD NORM")

        # Wait for trigger
        triggered = False
        wait_time = 0.0
        while not triggered and wait_time < 30.0:
            status = self.query("INR?")
            triggered = status == "INR 8193"
            wait_time += 0.25
            time.sleep(0.25)

        if not triggered:
            return None

        # Enable channel if needed
        channel_enabled = "ON" in self.query("C%d:TRA?" % n)
        if not channel_enabled:
            self.write("C%d:TRA ON" % n)
            time.sleep(3)

        self.write("WFSU SP,0,NP,0,FP,0")  # setup to get all data points

        def is_terminated(message):
            try:
                size = int(message[13:22])
            except ValueError:
                return False

            data_length = len(message[22:-2])

            if data_length == size:
                return True
            else:
                return False

        self.adapter.read_termination = is_terminated

        def validator(response):
            """
            Checks for correct header, size, message length and termination.
            """

            if response is None:
                return False
            elif len(response) == 0:
                return False

            header = response[:13]
            if header != (b"C%d:WF DAT2,#9" % n):
                return False

            size = response[13:22]
            try:
                size = int(size)
            except ValueError:
                return False

            data = response[22:-2]
            if len(data) != size:
                return False

            termination = response[-2:]
            if termination != b"\n\n":
                return False

            return True

        response = self.query("C%d:WF? DAT2" % n, decode=False, validator=validator)

        self.adapter.read_termination = "\n"  # restore normal read termination

        self.write("TRMD NORM")  # set to single trigger mode

        header, size, waveform = response[:13], response[13:22], response[22:-2]
        size = int(size)

        # convert bytes to integers
        waveform = np.array(struct.unpack("b" * size, waveform), dtype=np.float64)

        scale = self._get_chn_scale(n)
        pos = self._get_chn_position(n)

        waveform = (scale / 25.0) * waveform - pos

        return waveform

    @measurer
    def measure_ch1_waveform(self) -> Array:
        return self._measure_chn_waveform(1)

    @measurer
    def measure_ch2_waveform(self) -> Array:
        return self._measure_chn_waveform(2)

    @measurer
    def measure_ch3_waveform(self) -> Array:
        return self._measure_chn_waveform(3)

    @measurer
    def measure_ch4_waveform(self) -> Array:
        return self._measure_chn_waveform(4)

    @measurer
    def measure_sample_rate(self) -> Float:
        return float(self.query("SARA?")[5:-4])
